#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Image and Binary handlers. Provide also a Damona manager"""
import functools
import os
import pathlib
import re
import sys

import colorlog
from easydev import cmd_exists, md5

logger = colorlog.getLogger(__name__)


__all__ = ["Damona", "ImageReader", "BinaryReader", "DamonaInit"]


def get_damona_path():
    if "DAMONA_PATH" not in os.environ:
        logger.error(
            "DAMONA_PATH not found in your environment. You must define "
            "it. In this shell, type 'export DAMONA_PATH=PATH_WHERE_TO_PLACE_DAMONA'"
        )
        sys.exit(1)
    return pathlib.Path(os.environ["DAMONA_PATH"])


class DamonaInit:
    """Class to create images/bin directory for DAMONA


    This is called each time damona is started to make sure the
    required config file are present.

    This class simply create the *~/.config/damona/envs* and images directories.
    It also checks whether **DAMONA_PATH** and **DAMONA_SINGULARITY_OPTIONS** variables
    are defined in the environment.
    """

    def __init__(self):
        if "DAMONA_PATH" not in os.environ:
            logger.critical(
                """DAMONA_PATH was not found in your environment.

Before using Damona, you have to copy/paste the following code in
your ~/.bashrc file once for all (start a new shell afterwards):

    if [ ! -f  "~/.config/damona/damona.sh" ] ; then
        source ~/.config/damona/damona.sh
    fi

This will create DAMONA_PATH variable that points to your home/.config/damona/ directory.
You can redefine the DAMONA_PATH later to point towards another path if needed."""
            )
            # This is not an error per se but damona cannot work without DAMONA_PATH
            # Yet, we do not want to raise an error especially for the CI
            sys.exit(0)

        if "DAMONA_SINGULARITY_OPTIONS" not in os.environ:
            logger.warning(
                """No DAMONA_SINGULARITY_OPTIONS variable found in your environment.
To remove this message, set a DAMONA_SINGULARITY_OPTIONS variable in your shell. For explanation about
this variable, please see https://damona.readthedocs.io/en/latest/userguide.html#DAMONA_SINGULARITY_OPTIONS """
            )

        self.damona_path = pathlib.Path(os.environ["DAMONA_PATH"])
        os.makedirs(self.damona_path, exist_ok=True)
        os.makedirs(self.damona_path / "envs" / "base" / "bin", exist_ok=True)
        os.makedirs(self.damona_path / "images" / "damona_buffer", exist_ok=True)


class Damona:
    """Global manager to get information about environments, binaries, images."""

    def __init__(self):
        #: This attribute stored the path where images and environments are stored
        self.damona_path = get_damona_path()

    def _get_config_path(self):
        return self.damona_path / "damona.cfg"

    config_path = property(_get_config_path, doc="Get the Damona config file location")

    def _get_image_directory(self):
        return self.damona_path / "images"

    images_directory = property(_get_image_directory, doc="Get the Damona images directory location")

    def _get_environments_path(self):
        return self.damona_path / "envs"

    environments_path = property(_get_environments_path, doc="Get the Damona environments directory location")

    def find_orphan_binaries(self):
        """Find binaries in all environments that are orphans

        By orphans, we mean that their image is not present anymore for some
        reasons (e.g., users delete it manually).
        """
        binaries = self.get_all_binaries()
        orphans = []
        for x in binaries:
            br = BinaryReader(x)
            if br.is_image_available() is False:  # pragma: no cover
                logger.warning(f"{x} image is not available. This binary is an orphan")
                orphans.append(x)
        return orphans

    def get_environments(self):
        """return the list of environments names"""
        from damona.environ import Environ

        env = Environ()
        return env.environment_names

    def get_all_binaries(self):
        """Return list of all binaries in all environments"""
        from damona.environ import Environ

        env = Environ()
        binaries = [e.get_installed_binaries() for e in env.environments]
        binaries = set([x for y in binaries for x in y])
        return binaries

    def find_orphan_images(self):
        """Get images that have no binaries in any environments"""
        binaries = self.get_all_binaries()
        images = self.get_all_images()
        Nb = len(binaries)
        Ni = len(images)

        # keep print to make sure it is seen
        print(f"Found {Ni} images and {Nb} binaries. Checking consistencies")
        used_images = []
        for binary in binaries:
            br = BinaryReader(binary)
            used_images.append(pathlib.Path(br.image))

        used_images = set(used_images)
        Nu = len(used_images)
        # No = Ni - Nu
        # keep print to make sure it is seen
        print(f"{Nu} images is/are used. ")
        orphans = []

        for image in sorted(images):
            if image not in used_images:  # pragma: no cover
                logger.info(f"{image} image not used.")
                orphans.append(image)
        return orphans

    def get_all_images(self):
        """Return list of all images"""
        from damona.environ import Images

        images = Images()
        return list(images.files)

    def is_image_used(self, name):
        """Return True if this image is used

        The image name has no ".img" extension and uses _ instead of : character
        ::

            # get images used
            from damona import Damona
            d = Damona()
            d.get_all_images()

            # Note that names are encoded as NAME_X.Y.Z

            d.is_image_used("fastqc_0.11.9")
        """
        images_used = set([x for x in self.get_all_binaries() if name == BinaryReader(x).get_image().replace(":", "_")])
        return bool(images_used)


class ImageReader:
    """Manage a single Singularity image"""

    def __init__(self, name):
        """.. rubric:: **Constructor**

        :param name: the input name of the image (fullpath)

        ::

            >>> from damona.common import ImageReader
            >>> ir = ImageReader("~/.config/damona/images/fastqc_0.11.9.img")
            >>> ir.md5
            >>> ir.is_orphan()
            >>> ir.name
            >>> print(ir.shortname)
            'fastqc_0.11.9.img'
            >>> print(ir.version)
            '0.11.9'

        """
        self.filename = pathlib.Path(name)

        if self.is_valid_name() is False:
            logger.error(f"Invalid image name ({self.shortname}). Your input image must end in .img or .sif ; version must be X.Y.Z")
            sys.exit(1)

    def delete(self):
        if self.is_orphan():
            logger.warning(f"deleting {self.filename} since it is not used anymore by any environments")
            self.filename.unlink()
        else:
            logger.warning(
                f"{self.filename} not deleted because it is still used. Removing an image that is used is not yet implemented"
            )

    def _get_short_name(self):
        return self.filename.name

    shortname = property(_get_short_name, doc="Get the filename (NAME_X.Y.Z.img)")

    def is_valid_name(self):
        """Check whether the name is valid.


        Must be in the form NAME_X.Y.Z.img
        """
        pattern = r".+_(v|)\d+\.\d+\.\d+(.+|)\.(img|sif)"
        p = re.compile(pattern)
        if p.match(self.shortname):
            return True
        else:
            return False

    def _get_executable_name(self):
        pattern = r"_(v|)\d+\.\d+\.\d+(.+|)\.(img|sif)"
        p = re.compile(pattern)
        ss = p.search(self.shortname)
        guess = self.shortname[0 : ss.span()[0]]
        return guess

    guessed_executable = property(_get_executable_name, doc="Guess the executable from the filename")

    def _get_version(self):
        pattern = r"_(v|)\d+\.\d+\.\d+(.+|)\.(img|sif)"
        p = re.compile(pattern)
        ss = p.search(self.shortname)
        version = ss.group().replace(".sif", "").replace(".img", "")
        if version[0] == "_":
            version = version[1:]
        if version[0] == "v":
            version = version[1:]
        return version

    version = property(_get_version, doc="Get the version")

    def _get_md5sum(self):
        md5sum = md5(self.filename)
        return md5sum

    md5 = property(_get_md5sum, doc="compute and return the md5 of the file")

    def is_orphan(self):
        binaries = Damona().get_all_binaries()
        linked_binaries = []
        for binary in binaries:
            if BinaryReader(binary).is_image_available():
                linked_binaries.append(binary)

        if len(linked_binaries) == 0:
            return True
        else:
            return False

    def is_installed(self):
        """Return True is the file exists in the DAMONA_PATH"""
        damona_path = pathlib.Path(os.environ["DAMONA_PATH"])
        if (damona_path / "images" / self.filename.name).exists():
            return True
        else:
            return False

    def __repr__(self):
        txt = f"name: {self.filename.absolute()}\n"
        txt += f"shortname: {self.shortname}\n"
        txt += f"md5: {self.md5}\n"
        txt += f"version: {self.version}\n"
        txt += f"guessed executable: {self.guessed_executable}"
        return txt


class BinaryReader:
    """Manage a single binary

    ::

        >>> from damona.common import BinaryReader
        >>> br = BinaryReader("~/.config/damona/envs/base/bin/fastqc")
        >>> br.get_image()
        'fastqc:0.11.9'
        >>> br.is_image_available()
        True
    """

    def __init__(self, filename):
        """.. rubric:: constructor

        :param str filename: the input name of the binary file

        Can be use to check whether the binary is not orphan and its image is
        still available.

        """
        logger.debug(f"{filename}")
        if isinstance(filename, str):
            filename = pathlib.Path(filename)
        self.filename = filename

        with self.filename.open("r") as fin:
            data = [x for x in fin.readlines() if x.strip().startswith("singularity")]
            data = data[0]

            data = data.replace("${DAMONA_SINGULARITY_OPTIONS}", "")
            try:
                image_path = data.split("exec")[1].split()[0]
            except:  # pragma: no cover
                image_path = data.split("run")[1].split()[0]
                logger.warning(f"command line in {self.filename} uses 'run'; should be reinstalled ")

            if "DAMONA_PATH" in os.environ:
                DAMONA_PATH = os.environ["DAMONA_PATH"]
                self.image = image_path.replace("${DAMONA_PATH}", DAMONA_PATH)
            else:
                self.image = image_path

    def is_image_available(self):
        """Return True if the image used by the binary does exist"""
        if "DAMONA_PATH" not in os.environ:
            logger.error("You must define DAMONA_PATH")
            sys.exit(1)

        damona_path = os.environ["DAMONA_PATH"]
        if os.path.exists(self.image.replace("${DAMONA_PATH}", damona_path)):
            return True
        else:
            return False

    def get_image(self):
        """Return the container used by the binary"""
        # we assume the user did not edit the binary file
        # so we expect one uncommented line
        with self.filename.open("r") as fin:
            command = [line for line in fin.readlines() if line.strip() and line.strip()[0] != "#"]
        # where /images is to be followed by the container
        image = [x for x in command[0].split() if "/images/" in x]
        image = image[0].split()
        container = image[0].split("/")[-1]
        container = container.replace(".img", "")
        container = ":".join(container.rsplit("_", 1))
        return container


def requires_singularity(func):
    """A decorator to check presence of singularity"""

    @functools.wraps(func)
    def wrapper(ref, *args, **kwargs):
        if cmd_exists("singularity"):
            return func(ref, *args, **kwargs)
        else:
            logger.error("singularity command was not found. You must install 'singularity' to use Damona")

    return wrapper



#
#  This file is part of Damona software
#
#  Copyright (c) 2020 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Tools to upload/retrieve zenodo deposits"""
import inspect
import json
import os
import sys
from configparser import NoOptionError, NoSectionError

from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

import colorlog
import requests

from damona import Config
from damona.registry import ImageName, Software

logger = colorlog.getLogger(__name__)


logger.setLevel("INFO")


class Zenodo:  # pragma: no cover
    """

    ::

        z = Zenodo(mode="sandbox.zenodo")

    You can retrieve an existing deposit (read-only) given its ID::

        deposit = z.get_deposition(959590)

    Or start with a new one::

        deposit = z.create_new_deposition()


    If you want to create a new file deposition and publish it, follow those 4 commands:

    1. create a new deposit::

        deposit = z.create_new_deposition()
        ID = deposit['id']
        # e.g. 959590

    2. upload data::

        up = z.upload('file', deposit)
        # this gives us the md5sum for later

    3. metadata. Here, we update the deposit with some metadata::

        data = z.get_metadata("fastqc", "v0.11.8")
        update = z.update_metadata(deposit, data)

    4. publish  NOTE THAT PUBLISHED DEPOSITION CANNOT BE MODIFIED NOR DELETED::

        published = z.publish(deposit)

    # if we want an update, we must create a new version from a published deposit::

        new_version = z.create_new_deposit_version(deposit)


    A damona registry looks like::

        ivar:
          doi: 10.5281/zenodo.8033025
          zenodo_id: 8033026
          releases:
            1.3.1:
              download: https://zenodo.org/record/8033026/files/ivar_1.3.1.img
              md5sum: 213e286c708e12d6423b0bd8fa723327
              doi: 10.5281/zenodo.8033026
              filesize: 139108352

    This is created when using::

        damona zenodo-upload file.img --mode zenodo

    The zenodo_id is the record on the main page. e.g https://zenodo.org/records/8033026
    To cite all versions, use the main doi (here 10.5281/zenodo.8033025).

    Each release has its own DOI in case you want to cite a specific version.

    """

    def __init__(self, mode="sandbox.zenodo", token=None, author=None, affiliation=None, orcid=None):
        assert mode in ["zenodo", "sandbox.zenodo"]
        self.mode = mode
        self.headers = {"Content-Type": "application/json"}
        self.last_requests = []

        cfg = Config()

        if token is None:
            try:
                self.token = cfg.config.get(f"{mode}", "token")
                logger.info(f"Found token for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                logger.error("A token must be provided on command line or in your damona.cfg file")
                sys.exit(1)
        else:
            self.token = token

        # Do we have an orcid (optional)
        if orcid is None:
            try:
                self.orcid = cfg.config.get(f"{mode}", "orcid")
                logger.info(f"Found ORCID for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                pass
        else:
            self.orcid = orcid

        # Do we have a name ?
        if author is None:
            try:
                self.author = cfg.config.get(f"{mode}", "name").replace("'", "").replace('"', "")
                logger.info(f"Found Name {self.author} for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                pass
        else:
            self.author = author

        # Do we have an affiliation ?
        if affiliation is None:
            try:
                self.affiliation = cfg.config.get(f"{mode}", "affiliation").replace("'", "").replace('"', "")
                logger.info(f"Found Affiliation {self.affiliation} for {mode} in {cfg.config_file}")
            except (NoSectionError, NoOptionError):
                pass
        else:
            self.affiliation = affiliation

        if self.affiliation is None or self.author is None:
            logger.error("you must provide an author and affiliation")
            sys.exit(1)

    def _get_params(self):
        return {"access_token": self.token}

    params = property(_get_params)

    def _status(self, r, correct_codes):
        try:
            caller = inspect.stack()[1].function
        except:
            caller = ""

        self.last_requests.append(r)

        if r.status_code not in correct_codes:
            logger.warning(f"Status code: {r.status_code}; {r.reason}")
            try:
                output = f"{r.json()}"
                logger.warning(output)
            except:
                logger.warning("Could not figure out the erorr from json()")
            finally:

                logger.error(f"unexpected return code {r.status_code}")
                logger.error(f"Caller: {caller}")
                logger.error("Know issues:\n - Have you set a token in zenodo and provide the token in damona.cfg ? ")
                sys.exit(1)

    def create_new_deposition(self):  # pragma: no cover
        # expected status is 201

        logger.info("Creating a new deposit")
        url = f"https://{self.mode}.org/api/deposit/depositions"
        r = requests.post(url, params=self.params, json={}, headers=self.headers)
        self._status(r, [201])

        try:
            ID = self.last_requests[-1].json()["id"]
            logger.info(f"Created a new deposit with ID={ID}")
        except:
            raise
        return r

    def get_all_depositions(self):  # pragma: no cover
        """Return all depositions for the currently authenticated user"""
        url = f"https://{self.mode}.org/api/deposit/depositions"
        r = requests.get(url, params=self.params, json={})
        self._status(r, [200])
        return r

    def get_deposition(self, ID):  # pragma: no cover
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}"
        r = requests.get(url, params=self.params, json={})
        self._status(r, [200])
        return r

    def delete_deposition(self, ID):  # pragma: no cover
        """only unpublished deposit can be deleted.

        Normal status code is 204. If already deleted, sent 401
        """
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}"
        r = requests.delete(url, params=self.params, json={})
        self._status(r, [201])
        return r

    def upload(self, filename, json_deposit):  # pragma: no cover
        logger.info("Uploading file")

        try:
            bucket_url = json_deposit["links"]["bucket"]
        except:
            bucket_url = json_deposit.json()["links"]["bucket"]

        jsons = {}

        # Determine the total size of the file
        total_size = os.path.getsize(filename)


        # the upload it self may take a while
        with open(filename, "rb") as fp:
            basename = os.path.basename(filename)

            with tqdm(total=total_size, unit="B", unit_scale=True, unit_divisor=1024) as t:
                wrapped_file = CallbackIOWrapper(t.update, fp, "read")
                r = requests.put(f"{bucket_url}/{basename}", data=wrapped_file, params=self.params)


            self._status(r, [200, 201])
            return r

    def get_metadata(self, software, version):
        # check validity of the version
        assert len(version.split(".")) == 3
        items = version.split(".")
        if items[0][0] != "v":
            version = f"v{version}"

        data = {
            "metadata": {
                "title": f"Damona singularity image of {software} software",
                "upload_type": "physicalobject",
                "description": f"""Singularity image(s) of {software} software to be used and installed with damona
(<a href="https://damona.readthedocs.org">See https://damona.readthedocs.org</a>) for reproducible bioinformatics
analysis.""",
                "keywords": ["apptainer", "singularity", "damona", "bioinformatics", "reproducibility", "container"],
                "version": f"{version}",
                "communities": [{"identifier": "damona"}],
            }
        }

        if self.orcid:
            data["metadata"]["creators"] = [{"orcid": self.orcid, "name": self.author, "affiliation": self.affiliation}]
        else:
            data["metadata"]["creators"] = [{"name": self.author, "affiliation": self.affiliation}]

        return data

    def update_metadata(self, data, deposit):  # pragma: no cover
        logger.info("Uploading metadata")
        ID = self.get_id(deposit)
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}"
        r = requests.put(url, params=self.params, data=json.dumps(data), headers=self.headers)
        self._status(r, [200])
        return r

    def publish(self, deposit):  # pragma: no cover
        ID = self.get_id(deposit)
        logger.info(f"Publishing {ID}")
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}/actions/publish"
        r = requests.post(url, params=self.params, headers=self.headers)
        self._status(r, [202, 504])
        return r

    def unlock(self, deposit):  # pragma: no cover
        ID = self.get_id(deposit)
        logger.info(f"Unlocking {ID}")
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}/actions/edit"
        r = requests.post(url, params=self.params)
        self._status(r, [201, 403])
        return r

    def create_new_deposit_version(self, deposit):  # pragma: no cover
        # expected status 201
        logger.info(f"Creating a new version for record {deposit}. Please wait")
        ID = self.get_id(deposit)
        url = f"https://{self.mode}.org/api/deposit/depositions/{ID}/actions/newversion"
        r = requests.post(url, params=self.params)
        self._status(r, [201])
        return r

    def get_id(self, data):
        """
        data can be the output of a requests, in which case, we call to_json() and expect a 'id'
        key. Otherwise, it is alrady a json structure. If none works, most probably just an ID (integer).
        """
        try:
            return data.json()["id"]
        except:
            try:
                return data["id"]
            except:
                return data

    def _get_registry(self):
        if self.mode == "zenodo":
            return "registry.yaml"
        elif self.mode == "sandbox.zenodo":
            return "registry_sandbox.yaml"

    registry_name = property(_get_registry)

    def _upload(self, filename):
        data = ImageName(filename)
        software = Software(data.name)

        # sandbox has no registry in general, so no name, therefore we create
        # a new deposit.
        if software.name and self.mode != "sandbox.zenodo":
            logger.info("Software known, adding new version.")
            msg = self.create_new_version_with_file_and_publish(filename)
            print(msg)
            with open(self.registry_name, "a+") as fout:
                fout.write(msg)
        else:
            logger.info("Software not known, adding new deposit.")
            msg = self.create_new_deposit_with_file_and_publish(filename)
            print(msg)
            with open(self.registry_name, "w") as fout:
                fout.write(msg)

    def create_new_deposit_with_file_and_publish(self, filename):  # pragma: no cover
        """ """
        data = ImageName(filename)

        logger.info(f"Creating and Publising new deposit for {data.name}_{data.version}")
        deposit = self.create_new_deposition()

        up = self.upload(filename, deposit)
        metadata = self.get_metadata(data.name, data.version)
        self.update_metadata(metadata, deposit)

        # in order to obtain the DOI, we must publish first
        self.publish(deposit)

        if self.last_requests[-1].ok:
            json = self.last_requests[-1].json()
            msg = self._print_info_new_deposit(data, json)
            return msg

    def _print_info_new_deposit(self, data, json):
        # figure out the filename entryy we have just added
        entry = [x for x in json["files"] if x["filename"] == data.basename][0]

        zenodo_id = json["id"]
        doi = json["conceptdoi"]
        msg = f"{data.name}:\n"
        msg += f"  doi: {doi}\n"
        msg += f"  zenodo_id: {zenodo_id}\n"
        msg += "  releases:\n"

        this_doi = json["doi"]
        record_html = json["links"]["record_html"]

        md5sum = entry["checksum"]
        filesize = entry["filesize"]
        basename = entry["filename"]
        msg += f"    {data.version}:\n"
        msg += f"      download: {record_html}/files/{basename}\n"
        msg += f"      md5sum: {md5sum}\n"
        msg += f"      doi: {this_doi}\n"
        msg += f"      filesize: {filesize}\n"  # no EOF
        return msg

    def create_new_version_with_file_and_publish(self, filename, deposit=None):
        data = ImageName(filename)
        registry_data = Software(data.name)._data
        try:
            zenodo_id = registry_data[data.name]["zenodo_id"]
        except KeyError:
            logger.critical(
                f"Could not find {data.name} in the registry. Filename must have the same name as those to be found in the recipes."
            )
            sys.exit(1)

        logger.info(f"Creating and Publising new version for {data.name}_{data.version}")

        deposit = self.create_new_deposit_version(zenodo_id)

        # switch the state from 'done' to 'inprogress'
        self.unlock(deposit)

        # get the ID of the new latest draft and its deposition
        new_ID = deposit.json()["links"]["latest_draft"].split("/")[-1]
        new_deposit = self.get_deposition(new_ID)

        up = self.upload(filename, new_deposit.json())

        metadata = self.get_metadata(data.name, data.version)
        self.update_metadata(metadata, new_deposit)

        self.publish(new_ID)

        if self.last_requests[-1].ok:
            json = self.last_requests[-1].json()
            msg = self._print_info_new_version(data, json)
            return msg

    def _print_info_new_version(self, data, json):
        # figure out the filename entryy we have just added
        entry = [x for x in json["files"] if x["filename"] == data.basename][0]

        # update the new zenodo ID

        this_doi = json["doi"]
        md5sum = entry["checksum"]
        basename = entry["filename"]
        record_html = json["links"]["record_html"]
        filesize = entry["filesize"]

        msg = f"    {data.version}:\n"
        msg += f"      download: {record_html}/files/{basename}\n"
        msg += f"      md5sum: {md5sum}\n"
        msg += f"      doi: {this_doi}\n"
        msg += f"      filesize: {filesize}"  # no EOF
        return msg


def get_stats_software(software):
    """Returns number of downloads"""
    from damona.registry import Software

    s = Software(software)
    try:
        return get_stats_id(s.zenodo_id, software)
    except AttributeError:
        return 0


def get_stats_id(ID, name=None):
    """Returns number of downloads


    For instance, for this link: https://zenodo.org/record/7319782
    you should provide the ID 7319782


    """
    import json

    from bs4 import BeautifulSoup

    if ID == "bioconainers":
        return 0
    elif ID is None:
        print(f"# ID {ID} is unknown for {name}. developers should update registry")
        return 0
    else:

        r = requests.get(f"https://zenodo.org/record/{ID}")
        bs = BeautifulSoup(r.content, features="html.parser")

        # according to damona, there is a limit of 60 requests per minute but also
        # 2000 requests per hour. Since there are more than 60 software, we expect 
        # this call to reach the limit. Therefore, we introspect the X-RateLimit
        # and Number of requests remaining and add a sleep.
        import time
        R = int(r.headers['X-RateLimit-Remaining'])
        L = int(r.headers['X-RateLimit-Limit'])
        reset = int(r.headers['X-RateLimit-Reset'])
        T = int(time.time())
        if int(r.headers['X-RateLimit-Remaining']) < 1:

            delay = reset - int(time.time())
            logger.warning(f"Warning limit attained. please wait {delay}")
            time.sleep(delay)


        try:
            data = bs.find(id="recordVersions")
            data = json.loads(data.attrs["data-record"])["stats"]["all_versions"]["downloads"]
            return data
        except Exception as err:
            print(err)
            print(ID)
            print("====")
            return -1



#
#  This file is part of Damona software
#
#  Copyright (c) 2020 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
import functools
import pathlib
import shutil

import requests
from tqdm import tqdm


def download_with_progress(url, filename):
    """Download image/container from a URL.

    :param filename: where to save the container


    """
    resp = requests.get(url, stream=True, allow_redirects=True)
    if resp.status_code != 200:
        resp.raise_for_status()
        raise RuntimeError(f"Request to {url} returned status code {r.status_code}")

    file_size = int(resp.headers.get("Content-Length", 0))

    path = pathlib.Path(filename).expanduser().resolve()
    path.parent.mkdir(parents=True, exist_ok=True)

    desc = "(Unknown total file size)" if file_size == 0 else ""
    resp.raw.read = functools.partial(resp.raw.read, decode_content=True)  # Decompress if needed
    with tqdm.wrapattr(resp.raw, "read", total=file_size, desc=desc) as fout:
        with path.open("wb") as f:
            shutil.copyfileobj(fout, f)
    return path



#!/usr/bin/env python
""" MultiQC module to parse output from sequana"""
import os
import re

import logging
import json

logging.captureWarnings(True)
from multiqc import config
from multiqc.plots import linegraph, table, heatmap, bargraph
from multiqc.modules.base_module import BaseMultiqcModule

logging.captureWarnings(False)

# Initialise the logger
import colorlog

log = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/quality_control",
            anchor="sequana_quality_control",
            target="sequana_quality_control",
            href="http://github.com/sequana/sequana/",
            info="(sequana quality control multi summary)",
        )

        self.data = {}
        for myfile in self.find_log_files("sequana_quality_control"):
            try:
                thisdata = self.parse_logs(myfile["f"])
                name = thisdata["project"]
                self.data[name] = self.parse_logs(myfile["f"])
            except KeyError:
                pass

        if len(self.data) == 0:
            log.debug("No samples found: sequana_quality_control")
            raise UserWarning

        log.info("Found {} reports".format(len(self.data)))

        name = list(self.data.keys())[0]
        if "phix_section" in self.data[name]:
            self._has_phix = True
        else:
            self._has_phix = False

        self.populate_columns()
        self.add_phix_section()
        self.add_gc_section()
        self.add_adapter_section()

    def populate_columns(self):
        headers = {}
        if any(["multiqc_total_reads" in self.data[s] for s in self.data]):
            headers["multiqc_total_reads"] = {
                "title": "number of reads",
                "description": "number of reads",
                #'max': 100,
                "min": 0,
                #'modify': lambda x: x * 100,
                "scale": "RdYlGn",
                #'format': '{:,.1f}'
                "shared_key": "multiqc_total_reads",
                "hidden": True,
            }
        if self._has_phix:
            headers["phix"] = {"title": "Phix", "min": 0, "max": 100, "hidden": True}

        if len(headers.keys()):
            self.general_stats_addcols(self.data, headers)

    def parse_logs(self, log_dict):
        data = json.loads(log_dict)
        this = data["cutadapt_json"]["Number of reads"]["Total paired reads"]
        data["multiqc_total_reads"] = this
        # Phix removal is optional
        if "phix_section" in data.keys():
            this = data["phix_section"]["contamination"]
            data["phix"] = this
        return data

    def add_gc_section(self):
        data = {}
        for name in self.data.keys():
            data[name] = {"GC": self.data[name]["fastq_stats_samples_json"]["GC content"]["R1"]}
        pconfig = {"title": "Percentage of GC in the raw dat", "percentages": True, "min": 0, "max": 100}

        self.add_section(
            name="GC percent", anchor="GC percent", description="", helptext="", plot=bargraph.plot(data, None, pconfig)
        )

    def add_phix_section(self):
        pconfig = {"title": "Percentage of phix in the raw data", "percentages": True, "min": 0, "max": 100}
        if self._has_phix:
            data = {}
            for name in self.data.keys():
                data[name] = {"phix_qc": self.data[name]["phix_section"]["contamination"]}
            description = "Amount of Phix present in the raw data (and removed)"
            if sum([float(x["phix_qc"]) for x in data.values()]) == 0:
                plot = None
                description += "\n\nNo Phix found in the data."
            else:
                plot = bargraph.plot(data, None, pconfig)
        else:
            description = "No Phix removed (there may be some)"
            plot = None

        self.add_section(
            name="Phix presence", anchor="mean_read_length", description=description, helptext="", plot=plot
        )

    def add_adapter_section(self):
        data = {}
        for name in self.data.keys():
            thisdata = self.data[name]["cutadapt_json"]["percent"]["Pairs kept"]
            data[name] = {"pairs_kept": thisdata.replace("(", "").replace(")", "").replace("%", "")}

        pconfig = {
            "title": "Percentage of pairs kept",
            "percentages": True,
            "min": 0,
            "max": 100,
        }

        self.add_section(
            name="Pairs kept",
            anchor="Pairs kept",
            description="Pairs kept",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )



#!/usr/bin/env python

""" MultiQC module to parse output from sequana"""
import os
import re

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

# Initialise the logger
import colorlog

log = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/isoseq",  # name that appears at the top
            anchor="sequana_isoseq",  # ??
            target="sequana_isoseq",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="pipelines multi Summary",
        )

        self.sequana_data = {}
        for myfile in self.find_log_files("sequana_isoseq"):
            name = myfile["s_name"]

            try:
                parsed_data = self.parse_logs(myfile["f"])
            except Exception as err:
                print("{} {} could not be parsed".format(name, myfile["f"]))
                print(err)
                continue
            name = parsed_data["s_name"]
            self.sequana_data[name] = parsed_data

        if len(self.sequana_data) == 0:
            log.debug("No samples found: sequana_isoseq")
            raise UserWarning

        info = "<ul>"
        for this in sorted(self.sequana_data.keys()):
            info += '<li><a href="{}/summary.html">{}</a></li>'.format(this, this, this)
        info += "</ul>"
        href = "http://sequana.readthedocs.io/en/main/"
        target = "Sequana"
        mname = '<a href="{}" target="_blank">{}</a> individual report pages:'.format(href, target)
        self.intro = "<p>{} {}</p>".format(mname, info)

        log.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_ccs_reads_section()
        self.add_ccs_mean_length_section()
        self.add_isoforms("hq")
        self.add_isoforms("lq")
        self.add_polyA()
        self.add_primes("three")
        self.add_primes("five")
        self.add_flnc()
        # self.add_ratio_flnc_hq()

    def add_flnc(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"flnc": self.sequana_data[name]["flnc"]}
        pconfig = {
            "title": "flnc",
            "logswitch": False,
        }
        self.add_section(
            name="Number of flnc",
            anchor="flnc",
            description="flnc",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_primes(self, label):
        name2 = "{} primes".format(label)
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"3 {}".format(label): self.sequana_data[name]["{}_primes".format(label)]}
        pconfig = {
            "title": name2,
            "logswitch": True,
        }
        self.add_section(
            name="Number of {}".format(name2),
            anchor=name2,
            description=name2,
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_polyA(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"polyA": self.sequana_data[name]["polyA"]}
        pconfig = {
            "title": "polyA",
            "logswitch": True,
        }
        self.add_section(
            name="Number of polyA",
            anchor="poylA",
            description="polyA",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_isoforms(self, mode):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"number_{}_isoforms".format(mode): self.sequana_data[name]["number_{}_isoforms".format(mode)]}
        pconfig = {
            "title": "Number of {} isoforms".format(mode.upper()),
            "logswitch": True,
        }
        self.add_section(
            name="Number of {} isoforms".format(mode.upper()),
            anchor="number_{}_isoforms".format(mode),
            description="Number of {} isoforms".format(mode.upper()),
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_ccs_reads_section(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"number_ccs_reads": self.sequana_data[name]["number_ccs_reads"]}

        pconfig = {
            "title": "Number of CCS reads",
            "percentages": False,
            "min": 100,
            "logswitch": True,
        }
        self.add_section(
            name="Number of CCS reads",
            anchor="number_ccs_reads",
            description="Number of CCS reads",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_ccs_mean_length_section(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {
                "mean": self.sequana_data[name]["mean_length"],
            }
        pconfig = {
            "title": "Mean CCS read length",
            "percentages": False,
            "min": 100,
            "logswitch": True,
        }

        self.add_section(
            name="Mean CCS read length",
            anchor="mean_ccs_read_length",
            description="Mean CCS length of the reads",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def parse_logs(self, log_dict):
        import json

        log_dict = json.loads(log_dict)
        data = {}
        # data['count'] = log_dict['data']["count"]
        data["mean_length"] = log_dict["data"]["CCS"]["mean_length"]
        data["number_ccs_reads"] = log_dict["data"]["CCS"]["number_ccs_reads"]
        data["number_hq_isoforms"] = log_dict["data"]["hq_isoform"]["N"]
        data["number_lq_isoforms"] = log_dict["data"]["lq_isoform"]["N"]
        data["flnc"] = log_dict["data"]["classification"]["full_length_non_chimeric_reads"]
        data["polyA"] = log_dict["data"]["classification"]["polyA_reads"]
        data["three_primes"] = log_dict["data"]["classification"]["three_prime_reads"]
        data["five_primes"] = log_dict["data"]["classification"]["five_prime_reads"]
        # data["ratio_flnc_hq"] = data["flnc"] / data["number_hq_isoforms"]

        data["alldata"] = log_dict
        data["s_name"] = log_dict["sample_name"]

        return data

    def populate_columns(self):
        headers = {}
        if any(["mean_length" in self.sequana_data[s] for s in self.sequana_data]):
            headers["mean_length"] = {
                "title": "CCS mean read length",
                "description": "CCS mean read length",
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.0d}",
                "shared_key": "count",
            }

        if any(["number_ccs_reads" in self.sequana_data[s] for s in self.sequana_data]):
            headers["number_ccs_reads"] = {
                "title": "CCS reads",
                "description": "Number of CCS reads",
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.0d}",
            }

        for this in ["number_hq_isoforms", "number_lq_isoforms"]:
            if any([this in self.sequana_data[s] for s in self.sequana_data]):
                headers[this] = {
                    "title": " ".join(this.split()),
                    "description": " ".join(this.split()),
                    "min": 0,
                    "scale": "RdYlGn",
                    "format": "{:,.0d}",
                }
        if any(["polyA" in self.sequana_data[s] for s in self.sequana_data]):
            headers["polyA"] = {"title": "polyA", "description": "", "min": 0, "scale": "RdYlGn", "format": "{:,.0d}"}

        if any(["three_primes" in self.sequana_data[s] for s in self.sequana_data]):
            headers["three_primes"] = {
                "title": "three prime",
                "description": "",
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.0d}",
            }
        if any(["five_primes" in self.sequana_data[s] for s in self.sequana_data]):
            headers["five_primes"] = {
                "title": "five prime",
                "description": "",
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.0d}",
            }
        if any(["flnc" in self.sequana_data[s] for s in self.sequana_data]):
            headers["flnc"] = {"title": "flnc", "description": "", "min": 0, "scale": "RdYlGn", "format": "{:,.0d}"}
        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)



#
#  This file is part of Sequana software
#
#  Copyright (c) 2016-2021 - Sequana Dev Team (https://sequana.readthedocs.io)
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  Website:       https://github.com/sequana/sequana
#  Documentation: http://sequana.readthedocs.io
#  Contributors:  https://github.com/sequana/sequana/graphs/contributors
##############################################################################


import pandas as pd


class Reader:
    def __init__(self, filename):
        self.filename = filename
        self.reader()

    def reader(self):
        self.df = pd.read_csv(self.filename, sep="\t")


class Bowtie1Reader(Reader):
    def __init__(self, filename):
        super().__init__(filename)

    def plot_bar(self, html_code=False):
        import plotly.graph_objects as go

        fig = go.Figure()

        x = self.df["not_aligned_percentage"]
        compx = 100 - x

        fig.add_trace(go.Bar(y=self.df["Sample"], x=compx, name="Aligned", marker_color="#389f1f", orientation="h"))
        fig.add_trace(
            go.Bar(
                y=self.df.Sample,
                x=self.df["not_aligned_percentage"],
                name="Not Aligned",
                marker_color="#120946",
                orientation="h",
            )
        )

        # Here we modify the tickangle of the xaxis, resulting in rotated labels.
        fig.update_layout(barmode="stack", xaxis_tickangle=-45, height=400, title="Mapping on ribosomal/contaminant")

        if html_code:
            return fig
        else:
            fig.show()


class Bowtie2(Reader):
    def __init__(self, filename):
        super().__init__(filename)

    def plot(self, html_code=False):
        import plotly.graph_objects as go

        fig = go.Figure()

        # get percentage instead of counts
        if "SE mapped uniquely" in self.df:

            S = self.df[["SE mapped uniquely", "SE multimapped", "SE not aligned"]].sum(axis=1).values
            df = self.df[["SE mapped uniquely", "SE multimapped", "SE not aligned"]].divide(S, axis=0) * 100
            df["Sample"] = self.df["Sample"]

            fig = go.Figure(
                data=[
                    go.Bar(
                        name="SE mapped uniquely",
                        y=df.Sample,
                        orientation="h",
                        x=df["SE mapped uniquely"],
                        marker_color="#17478f",
                    ),
                    go.Bar(
                        name="SE multimapped",
                        y=df.Sample,
                        orientation="h",
                        x=df["SE multimapped"],
                        marker_color="#e2780d",
                    ),
                    go.Bar(
                        name="SE not alinged",
                        y=df.Sample,
                        orientation="h",
                        x=df["SE not aligned"],
                        marker_color="#9f1416",
                    ),
                ]
            )
        else:
            columns = [
                "PE mapped uniquely",
                "PE mapped discordantly uniquely",
                "PE one mate mapped uniquely",
                "PE multimapped",
                "PE one mate multimapped",
                "PE neither mate aligned",
            ]
            colors = ["#20568f", "#5c94ca", "#95ceff", "#f7a35c", "#ffeb75", "#981919"]

            S = self.df[columns].sum(axis=1).values
            df = self.df[columns].divide(S, axis=0) * 100
            df["Sample"] = self.df["Sample"]

            fig = go.Figure(
                data=[
                    go.Bar(
                        name=column,
                        y=df.Sample,
                        orientation="h",
                        x=df[column],
                        marker_color=color,
                    )
                    for column, color in zip(columns, colors)
                ],
                layout_xaxis_range=[0, 100],
            )

        fig.update_layout(barmode="stack", title="", xaxis_title="Mapping rate (percentage)")

        if html_code:
            return fig
        else:
            fig.show()


class STAR(Reader):
    def __init__(self, filename):
        super().__init__(filename)

    def plot(self, html_code=False):
        import plotly.graph_objects as go

        fig = go.Figure()

        # get percentage

        df = self.df[["Sample"] + [x for x in self.df.columns if "_percent" in x]]

        fig = go.Figure(
            data=[
                go.Bar(
                    name="uniquely mapped",
                    y=df.Sample,
                    orientation="h",
                    x=df["uniquely_mapped_percent"],
                    marker_color="#437bb1",
                ),
                go.Bar(
                    name="Mapped to multiple loci",
                    y=df.Sample,
                    orientation="h",
                    x=df["multimapped_percent"],
                    marker_color="#7cb5ec",
                ),
                go.Bar(
                    name="Mapped to too many loci",
                    y=df.Sample,
                    orientation="h",
                    x=df["multimapped_toomany_percent"],
                    marker_color="#f7a35c",
                ),
                go.Bar(
                    name="Unmapped: to many mismatches",
                    y=df.Sample,
                    orientation="h",
                    x=df["unmapped_mismatches_percent"],
                    marker_color="#e63491",
                ),
                go.Bar(
                    name="Unmapped: too short",
                    y=df.Sample,
                    orientation="h",
                    x=df["unmapped_tooshort_percent"],
                    marker_color="#b1084c",
                ),
                go.Bar(
                    name="Unmapped: other",
                    y=df.Sample,
                    orientation="h",
                    x=df["unmapped_other_percent"],
                    marker_color="#028ce2",
                ),
            ]
        )

        fig.update_layout(barmode="stack", title="Alignment Scores", xaxis_title="Mapping rate (percentage)")

        if html_code:
            return fig
        else:
            fig.show()


class FeatureCounts(Reader):
    def __init__(self, filename):
        super().__init__(filename)

    def plot(self, html_code=False):
        import plotly.graph_objects as go

        fig = go.Figure()

        columns = ["Assigned", "Unassigned: Unmapped", "Unassigned: No Features", "Unassigned: Ambiguity"]
        S = self.df[columns].sum(axis=1).values
        df = self.df[columns].divide(S, axis=0) * 100
        df["Sample"] = self.df["Sample"]

        fig = go.Figure(
            data=[
                go.Bar(
                    name="Assigned",
                    y=df.Sample,
                    orientation="h",
                    x=df["Assigned"],
                    marker_color="#7cb5ec",
                ),
                go.Bar(
                    name="Unassigned: Unmapped",
                    y=df.Sample,
                    orientation="h",
                    x=df["Unassigned: Unmapped"],
                    marker_color="#434348",
                ),
                go.Bar(
                    name="Unassigned: No Features",
                    y=df.Sample,
                    orientation="h",
                    x=df["Unassigned: No Features"],
                    marker_color="#90ed7d",
                ),
                go.Bar(
                    name="Unassigned: Ambiguity",
                    y=df.Sample,
                    orientation="h",
                    x=df["Unassigned: Ambiguity"],
                    marker_color="#f7a35c",
                ),
            ]
        )

        fig.update_layout(
            barmode="stack", title="FeatureCounts: Assignments", xaxis_title="Annotation rate (percentage)"
        )

        if html_code:
            return fig
        else:
            fig.show()



import gzip
import io
import math
import os
import sys
from collections import OrderedDict, defaultdict, namedtuple
from itertools import chain

import numpy
from scipy.stats.stats import rankdata


def mean(items):
    items = list(items)
    return sum(items) / float(len(items))


import idr
import idr.optimization
from idr.optimization import estimate_model_params, old_estimator
from idr.utility import calc_post_membership_prbs, compute_pseudo_values

Peak = namedtuple("Peak", ["chrm", "strand", "start", "stop", "signal", "summit", "signalValue", "pValue", "qValue"])
MergedPeak = namedtuple("Peak", ["chrm", "strand", "start", "stop", "summit", "merged_signal", "signals", "pks"])


def load_gff(fp):
    """
    chr20   GRIT    TSS     36322438        36322468        44      +       .       gene_id 'chr20_plus_36322407_36500530'; gene_name 'chr20_plus_36322407_36500530'; tss_id 'TSS_chr20_plus_36322407_36500530_pk1'; peak_cov '7,0,11,0,0,0,0,0,3,0,1,0,0,0,6,0,0,0,0,0,3,0,4,0,0,0,8,0,0,1';
    """
    grpd_peaks = defaultdict(list)
    for line in fp:
        if line.startswith("#"):
            continue
        if line.startswith("track"):
            continue
        data = line.split()
        signal = float(data[5])
        peak = Peak(data[0], data[6], int(float(data[3])), int(float(data[4])), signal, None, None, None, None)
        grpd_peaks[(peak.chrm, peak.strand)].append(peak)
    return grpd_peaks


def load_bed(fp, signal_index, peak_summit_index=None):
    grpd_peaks = defaultdict(list)
    for line in fp:
        if line.startswith("#"):
            continue
        if line.startswith("track"):
            continue
        data = line.split()
        signal = float(data[signal_index])
        if idr.ONLY_ALLOW_NON_NEGATIVE_VALUES and signal < 0:
            raise ValueError("Invalid Signal Value: {:e}".format(signal))
        if peak_summit_index == None or int(data[peak_summit_index]) == -1:
            summit = None
        else:
            summit = int(data[peak_summit_index]) + int(float(data[1]))
        assert summit == None or summit >= 0
        peak = Peak(
            data[0],
            data[5],
            int(float(data[1])),
            int(float(data[2])),
            signal,
            summit,
            float(data[6]),
            float(data[7]),
            float(data[8]),
        )
        grpd_peaks[(peak.chrm, peak.strand)].append(peak)
    return grpd_peaks


def correct_multi_summit_peak_IDR_values(idr_values, merged_peaks):
    assert len(idr_values) == len(merged_peaks)
    new_values = idr_values.copy()
    # find the maximum IDR value for each peak
    pk_idr_values = defaultdict(lambda: float("inf"))
    for i, pk in enumerate(merged_peaks):
        pk_idr_values[(pk.chrm, pk.strand, pk.start, pk.stop)] = min(
            pk_idr_values[(pk.chrm, pk.strand, pk.start, pk.stop)], idr_values[i]
        )
    # store the indices best peak indices, and update the values
    best_indices = []
    for i, pk in enumerate(merged_peaks):
        region = (pk.chrm, pk.strand, pk.start, pk.stop)
        if new_values[i] == pk_idr_values[region]:
            best_indices.append(i)
        else:
            new_values[i] = pk_idr_values[region]
    return numpy.array(best_indices), new_values


def iter_merge_grpd_intervals(intervals, n_samples, pk_agg_fn, use_oracle_pks, use_nonoverlapping_peaks):
    # grp peaks by their source, and calculate the merged
    # peak boundaries
    grpd_peaks = OrderedDict([(i + 1, []) for i in range(n_samples)])
    pk_start, pk_stop = 1e12, -1
    for interval, sample_id in intervals:
        # if we've provided a unified peak set, ignore any intervals that
        # don't contain it for the purposes of generating the merged list
        if (not use_oracle_pks) or sample_id == 0:
            pk_start = min(interval.start, pk_start)
            pk_stop = max(interval.stop, pk_stop)
        # if this is an actual sample (ie not a merged peaks)
        if sample_id > 0:
            grpd_peaks[sample_id].append(interval)

    # if there are no identified peaks, continue (this can happen if
    # we have a merged peak list but no merged peaks overlap sample peaks)
    if pk_stop == -1:
        return None

    # skip regions that dont have a peak in all replicates
    if not use_nonoverlapping_peaks:
        if any(0 == len(peaks) for peaks in grpd_peaks.values()):
            return None

    # find the merged peak summit
    # note that we can iterate through the values because
    # grpd_peaks is an ordered dict
    replicate_summits = []
    for sample_id, pks in grpd_peaks.items():
        # if an oracle peak set is specified, skip the replicates
        if use_oracle_pks and sample_id != 0:
            continue

        # initialize the summit to the first peak
        try:
            replicate_summit, summit_signal = pks[0].summit, pks[0].signal
        except IndexError:
            replicate_summit, summit_signal = None, -1e9
        # if there are more peaks, take the summit that corresponds to the
        # replicate peak with the highest signal value
        for pk in pks[1:]:
            if pk.summit != None and pk.signal > summit_signal:
                replicate_summit, summit_signal = pk.summit, pk.signal
        # make sure a peak summit was specified
        if replicate_summit != None:
            replicate_summits.append(replicate_summit)

    summit = int(mean(replicate_summits)) if len(replicate_summits) > 0 else None

    # note that we can iterate through the values because
    # grpd_peaks is an ordered dict
    signals = [pk_agg_fn(pk.signal for pk in pks) if len(pks) > 0 else 0 for pks in grpd_peaks.values()]
    merged_pk = (pk_start, pk_stop, summit, pk_agg_fn(signals), signals, grpd_peaks)

    yield merged_pk
    return


def iter_matched_oracle_pks(pks, n_samples, pk_agg_fn, use_nonoverlapping_peaks=False):
    """Match each oracle peak to it nearest replicate peaks."""
    oracle_pks = [pk for pk, sample_id in pks if sample_id == 0]
    # if there are zero oracle peaks in this
    if len(oracle_pks) == 0:
        return None
    # for each oracle peak, find score the replicate peaks
    for oracle_pk in oracle_pks:
        peaks_and_scores = OrderedDict([(i + 1, []) for i in range(n_samples)])
        for pk, sample_id in pks:
            # skip oracle peaks
            if sample_id == 0:
                continue

            # calculate the distance between summits, setting it to a large
            # value in case the peaks dont have summits
            summit_distance = sys.maxsize
            if oracle_pk.summit != None and pk.summit != None:
                summit_distance = abs(oracle_pk.summit - pk.summit)
            # calculate the fraction overlap witht he oracle peak
            overlap = 1 + min(oracle_pk.stop, pk.stop) - max(oracle_pk.start, pk.start)
            overlap_frac = overlap / (oracle_pk.stop - oracle_pk.start + 1)

            peaks_and_scores[sample_id].append(((summit_distance, -overlap_frac, -pk.signal), pk))

        # skip regions that dont have a peak in all replicates.
        if not use_nonoverlapping_peaks and any(0 == len(peaks) for peaks in peaks_and_scores.values()):
            continue

        # build the aggregated signal value, which is jsut the signal value
        # of the replicate peak witgh the closest match
        signals = []
        rep_pks = []
        for rep_id, scored_pks in peaks_and_scores.items():
            scored_pks.sort()
            if len(scored_pks) == 0:
                assert use_nonoverlapping_peaks
                signals.append(0)
                rep_pks.append(None)
            else:
                signals.append(scored_pks[0][1].signal)
                rep_pks.append(
                    [
                        scored_pks[0][1],
                    ]
                )
        all_peaks = [
            oracle_pk,
        ] + rep_pks
        new_pk = (
            oracle_pk.start,
            oracle_pk.stop,
            oracle_pk.summit,
            pk_agg_fn(signals),
            signals,
            OrderedDict(zip(range(len(all_peaks)), all_peaks)),
        )
        yield new_pk

    return


def merge_peaks_in_contig(all_s_peaks, pk_agg_fn, oracle_pks=None, use_nonoverlapping_peaks=False):
    """Merge peaks in a single contig/strand.

    returns: The merged peaks.
    """
    # merge and sort all peaks, keeping track of which sample they originated in
    oracle_pks_iter = []
    if oracle_pks != None:
        oracle_pks_iter = oracle_pks

    # merge and sort all of the intervals, leeping track of their source
    all_intervals = []
    for sample_id, peaks in enumerate(
        [
            oracle_pks_iter,
        ]
        + all_s_peaks
    ):
        all_intervals.extend((pk, sample_id) for pk in peaks)
    all_intervals.sort()

    # grp overlapping intervals. Since they're already sorted, all we need
    # to do is check if the current interval overlaps the previous interval
    grpd_intervals = [
        [],
    ]
    curr_start, curr_stop = all_intervals[0][:2]
    for pk, sample_id in all_intervals:
        if pk.start < curr_stop:
            curr_stop = max(pk.stop, curr_stop)
            grpd_intervals[-1].append((pk, sample_id))
        else:
            curr_start, curr_stop = pk.start, pk.stop
            grpd_intervals.append(
                [
                    (pk, sample_id),
                ]
            )

    # build the unified peak list, setting the score to
    # zero if it doesn't exist in both replicates
    merged_pks = []
    if oracle_pks == None:
        for intervals in grpd_intervals:
            for merged_pk in iter_merge_grpd_intervals(
                intervals,
                len(all_s_peaks),
                pk_agg_fn,
                use_oracle_pks=(oracle_pks != None),
                use_nonoverlapping_peaks=use_nonoverlapping_peaks,
            ):
                merged_pks.append(merged_pk)
    else:
        for intervals in grpd_intervals:
            for merged_pk in iter_matched_oracle_pks(
                intervals, len(all_s_peaks), pk_agg_fn, use_nonoverlapping_peaks=use_nonoverlapping_peaks
            ):
                merged_pks.append(merged_pk)

    return merged_pks


def merge_peaks(all_s_peaks, pk_agg_fn, oracle_pks=None, use_nonoverlapping_peaks=False):
    """Merge peaks over all contig/strands"""
    # if we have reference peaks, use its contigs: otherwise use
    # the union of the replicates contigs
    if oracle_pks != None:
        contigs = sorted(oracle_pks.keys())
    else:
        contigs = sorted(set(chain(*[list(s_peaks.keys()) for s_peaks in all_s_peaks])))

    merged_peaks = []
    for key in contigs:
        # check to see if we've been provided a peak list and, if so,
        # pass it down. If not, set the oracle peaks to None so that
        # the callee knows not to use them
        if oracle_pks != None:
            contig_oracle_pks = oracle_pks[key]
        else:
            contig_oracle_pks = None

        # since s*_peaks are default dicts, it will never raise a key error,
        # but instead return an empty list which is what we want
        merged_contig_peaks = merge_peaks_in_contig(
            [s_peaks[key] for s_peaks in all_s_peaks],
            pk_agg_fn,
            contig_oracle_pks,
            use_nonoverlapping_peaks=use_nonoverlapping_peaks,
        )
        merged_peaks.extend(MergedPeak(*(key + pk)) for pk in merged_contig_peaks)

    merged_peaks.sort(key=lambda x: x.merged_signal, reverse=True)
    return merged_peaks


def build_rank_vectors(merged_peaks):
    # allocate memory for the ranks vector
    s1 = numpy.zeros(len(merged_peaks))
    s2 = numpy.zeros(len(merged_peaks))
    # add the signal
    for i, x in enumerate(merged_peaks):
        s1[i], s2[i] = x.signals

    rank1 = numpy.lexsort((numpy.random.random(len(s1)), s1)).argsort()
    rank2 = numpy.lexsort((numpy.random.random(len(s2)), s2)).argsort()

    return (numpy.array(rank1, dtype=numpy.int), numpy.array(rank2, dtype=numpy.int))


def build_idr_output_line_with_bed6(m_pk, IDR, localIDR, output_file_type, signal_type, use_oracle_peak_values=True):
    # initialize the line with the bed6 entires - these are
    # present in all of the output types
    rv = [
        m_pk.chrm,
        str(m_pk.start),
        str(m_pk.stop),
        ".",
        "%i" % (min(1000, int(-125 * math.log2(IDR + 1e-12)))),
        m_pk.strand,
    ]
    if output_file_type == "bed":
        # if we just want a bed, there's nothing else to be done
        pass
    # for narrow/broad peak files, we need to add the 3 score fields
    elif output_file_type in ("narrowPeak", "broadPeak"):
        # if we want to use the oracle peak values for the scores, and an oracle
        # peak is specified
        if use_oracle_peak_values and 0 in m_pk.pks:
            signal_values = [m_pk.pks[0].signalValue, m_pk.pks[0].pValue, m_pk.pks[0].qValue]
            signal_values = ["%.5f" % x for x in signal_values]
        else:
            # set the signal values that we didn't use to -1 per the standard
            signal_values = ["-1", "-1", "-1"]
            signal_values[{"signal.value": 0, "p.value": 1, "q.value": 2}[signal_type]] = "%.5f" % m_pk.merged_signal
        rv.extend(signal_values)
        # if this is a narrow peak, we also need to add the summit
        if output_file_type == "narrowPeak":
            rv.append(str(-1 if m_pk.summit == None else m_pk.summit - m_pk.start))
    else:
        raise ValueError("Unrecognized output format '{}'".format(outputFormat))

    rv.append("%f" % -math.log10(max(1e-5, localIDR)))
    rv.append("%f" % -math.log10(max(1e-5, IDR)))

    for key, signal in enumerate(m_pk.signals):
        # we add one to the key because key=0 corresponds to the oracle peaks
        key += 1
        # if there is no matching peak for this replicate
        if m_pk.pks[key] is None:
            rv.append("-1")
            rv.append("-1")
            rv.append("%.5f" % signal)
            if output_file_type == "narrowPeak":
                rv.append("-1")
        else:
            rv.append("%i" % min(x.start for x in m_pk.pks[key]))
            rv.append("%i" % max(x.stop for x in m_pk.pks[key]))
            rv.append("%.5f" % signal)
            if output_file_type == "narrowPeak":
                rv.append("%i" % int(mean(x.summit - x.start for x in m_pk.pks[key])))

    return "\t".join(rv)


def build_backwards_compatible_idr_output_line(m_pk, IDR, localIDR, output_file_type, signal_type):
    rv = [
        m_pk.chrm,
    ]
    for key, signal in enumerate(m_pk.signals):
        rv.append("%i" % min(x.start for x in m_pk.pks[key + 1]))
        rv.append("%i" % max(x.stop for x in m_pk.pks[key + 1]))
        rv.append(".")
        rv.append("%.5f" % signal)

    rv.append("%.5f" % localIDR)
    rv.append("%.5f" % IDR)
    rv.append(m_pk.strand)

    return "\t".join(rv)


def calc_local_IDR(theta, r1, r2):
    """
    idr <- 1 - e.z
    o <- order(idr)
    idr.o <- idr[o]
    idr.rank <- rank(idr.o, ties.method = "max")
    top.mean <- function(index, x) {
        mean(x[1:index])
    }
    IDR.o <- sapply(idr.rank, top.mean, idr.o)
    IDR <- idr
    IDR[o] <- IDR.o
    """
    mu, sigma, rho, p = theta
    z1 = compute_pseudo_values(r1, mu, sigma, p, EPS=1e-12)
    z2 = compute_pseudo_values(r2, mu, sigma, p, EPS=1e-12)
    localIDR = 1 - calc_post_membership_prbs(numpy.array(theta), z1, z2)
    if idr.FILTER_PEAKS_BELOW_NOISE_MEAN:
        localIDR[z1 + z2 < 0] = 1

    # it doesn't make sense for the IDR values to be smaller than the
    # optimization tolerance
    localIDR = numpy.clip(localIDR, idr.CONVERGENCE_EPS_DEFAULT, 1)
    return localIDR


def calc_global_IDR(localIDR):
    local_idr_order = localIDR.argsort()
    ordered_local_idr = localIDR[local_idr_order]
    ordered_local_idr_ranks = rankdata(ordered_local_idr, method="max")
    IDR = []
    for i, rank in enumerate(ordered_local_idr_ranks):
        IDR.append(ordered_local_idr[:rank].mean())
    IDR = numpy.array(IDR)[local_idr_order.argsort()]
    return IDR


def fit_model_and_calc_local_idr(
    r1,
    r2,
    starting_point=None,
    max_iter=idr.MAX_ITER_DEFAULT,
    convergence_eps=idr.CONVERGENCE_EPS_DEFAULT,
    fix_mu=False,
    fix_sigma=False,
):
    # in theory we would try to find good starting point here,
    # but for now just set it to somethign reasonable
    if starting_point is None:
        starting_point = (idr.DEFAULT_MU, idr.DEFAULT_SIGMA, idr.DEFAULT_RHO, idr.DEFAULT_MIX_PARAM)

    idr.log("Initial parameter values: [%s]" % " ".join("%.2f" % x for x in starting_point))

    # fit the model parameters
    idr.log("Fitting the model parameters", "VERBOSE")
    if idr.PROFILE:
        import cProfile

        cProfile.runctx(
            """theta, loss = estimate_model_params(
                                    r1,r2,
                                    starting_point,
                                    max_iter=max_iter,
                                    convergence_eps=convergence_eps,
                                    fix_mu=fix_mu, fix_sigma=fix_sigma)
                                   """,
            {"estimate_model_params": estimate_model_params},
            {
                "r1": r1,
                "r2": r2,
                "starting_point": starting_point,
                "max_iter": max_iter,
                "convergence_eps": convergence_eps,
                "fix_mu": fix_mu,
                "fix_sigma": fix_sigma,
            },
        )
        assert False
    theta, loss = estimate_model_params(
        r1, r2, starting_point, max_iter=max_iter, convergence_eps=convergence_eps, fix_mu=fix_mu, fix_sigma=fix_sigma
    )

    idr.log("Finished running IDR on the datasets", "VERBOSE")
    idr.log("Final parameter values: [%s]" % " ".join("%.2f" % x for x in theta))

    # calculate the global IDR
    localIDRs = calc_local_IDR(numpy.array(theta), r1, r2)
    return localIDRs


def write_results_to_file(
    merged_peaks,
    output_file,
    output_file_type,
    signal_type,
    max_allowed_idr=1.0,
    soft_max_allowed_idr=1.0,
    localIDRs=None,
    IDRs=None,
    useBackwardsCompatibleOutput=False,
):
    if useBackwardsCompatibleOutput:
        build_idr_output_line = build_backwards_compatible_idr_output_line
    else:
        build_idr_output_line = build_idr_output_line_with_bed6

    # write out the result
    idr.log("Writing results to file", "VERBOSE")

    if localIDRs is None or IDRs is None:
        assert IDRs is None
        assert localIDRs is None
        localIDRs = numpy.ones(len(merged_peaks))
        IDRs = numpy.ones(len(merged_peaks))

    num_peaks_passing_hard_thresh = 0
    num_peaks_passing_soft_thresh = 0
    for localIDR, IDR, merged_peak in zip(localIDRs, IDRs, merged_peaks):
        # skip peaks with global idr values below the threshold
        if max_allowed_idr != None and IDR > max_allowed_idr:
            continue
        num_peaks_passing_hard_thresh += 1
        if IDR <= soft_max_allowed_idr:
            num_peaks_passing_soft_thresh += 1
        opline = build_idr_output_line(merged_peak, IDR, localIDR, output_file_type, signal_type)
        print(opline, file=output_file)

    if len(merged_peaks) == 0:
        return

    idr.log(
        "Number of reported peaks - {}/{} ({:.1f}%)\n".format(
            num_peaks_passing_hard_thresh,
            len(merged_peaks),
            100 * float(num_peaks_passing_hard_thresh) / len(merged_peaks),
        )
    )

    idr.log(
        "Number of peaks passing IDR cutoff of {} - {}/{} ({:.1f}%)\n".format(
            soft_max_allowed_idr,
            num_peaks_passing_soft_thresh,
            len(merged_peaks),
            100 * float(num_peaks_passing_soft_thresh) / len(merged_peaks),
        )
    )

    return


def parse_args():
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""
Program: IDR (Irreproducible Discovery Rate)
Version: {PACKAGE_VERSION}
Contact: Nathan Boley <npboley@gmail.com>
""".format(
            PACKAGE_VERSION=idr.__version__
        ),
    )

    def PossiblyGzippedFile(fname):
        if fname.endswith(".gz"):
            return io.TextIOWrapper(gzip.open(fname, "rb"))
        else:
            return open(fname, "r")

    parser.add_argument(
        "--samples", "-s", type=PossiblyGzippedFile, nargs=2, required=True, help="Files containing peaks and scores."
    )
    parser.add_argument(
        "--peak-list", "-p", type=PossiblyGzippedFile, help="If provided, all peaks will be taken from this file."
    )
    parser.add_argument(
        "--input-file-type",
        default="narrowPeak",
        choices=["narrowPeak", "broadPeak", "bed", "gff"],
        help="File type of --samples and --peak-list.",
    )

    parser.add_argument(
        "--rank",
        help="Which column to use to rank peaks."
        + "\t\nOptions: signal.value p.value q.value columnIndex"
        + "\nDefaults:\n\tnarrowPeak/broadPeak: signal.value\n\tbed: score",
    )

    default_ofname = "idrValues.txt"
    parser.add_argument(
        "--output-file",
        "-o",
        default=default_ofname,
        help="File to write output to.\nDefault: {}".format(default_ofname),
    )
    parser.add_argument(
        "--output-file-type",
        choices=["narrowPeak", "broadPeak", "bed"],
        default=None,
        help="Output file type. Defaults to input file type when available, otherwise bed.",
    )

    parser.add_argument(
        "--log-output-file",
        "-l",
        type=argparse.FileType("w"),
        default=sys.stderr,
        help="File to write output to. Default: stderr",
    )

    parser.add_argument(
        "--idr-threshold",
        "-i",
        type=float,
        default=None,
        help="Only return peaks with a global idr threshold below this value." + "\nDefault: report all peaks",
    )
    parser.add_argument(
        "--soft-idr-threshold",
        type=float,
        default=None,
        help="Report statistics for peaks with a global idr below this "
        + "value but return all peaks with an idr below --idr.\nDefault: %.2f" % idr.DEFAULT_SOFT_IDR_THRESH,
    )

    parser.add_argument("--use-old-output-format", action="store_true", default=False, help="Use old output format.")

    parser.add_argument("--plot", action="store_true", default=False, help="Plot the results to [OFNAME].png")

    parser.add_argument(
        "--use-nonoverlapping-peaks",
        action="store_true",
        default=False,
        help="Use peaks without an overlapping match and set the value to 0.",
    )

    parser.add_argument(
        "--peak-merge-method",
        choices=["sum", "avg", "min", "max"],
        default=None,
        help="Which method to use for merging peaks.\n"
        + "\tDefault: 'sum' for signal/score/column indexes, 'min' for p/q-value.",
    )

    parser.add_argument(
        "--initial-mu", type=float, default=idr.DEFAULT_MU, help="Initial value of mu. Default: %.2f" % idr.DEFAULT_MU
    )
    parser.add_argument(
        "--initial-sigma",
        type=float,
        default=idr.DEFAULT_SIGMA,
        help="Initial value of sigma. Default: %.2f" % idr.DEFAULT_SIGMA,
    )
    parser.add_argument(
        "--initial-rho",
        type=float,
        default=idr.DEFAULT_RHO,
        help="Initial value of rho. Default: %.2f" % idr.DEFAULT_RHO,
    )
    parser.add_argument(
        "--initial-mix-param",
        type=float,
        default=idr.DEFAULT_MIX_PARAM,
        help="Initial value of the mixture params. Default: %.2f" % idr.DEFAULT_MIX_PARAM,
    )

    parser.add_argument("--fix-mu", action="store_true", help="Fix mu to the starting point and do not let it vary.")
    parser.add_argument(
        "--fix-sigma", action="store_true", help="Fix sigma to the starting point and do not let it vary."
    )

    parser.add_argument(
        "--dont-filter-peaks-below-noise-mean",
        default=False,
        action="store_true",
        help="Allow signal points that are below the noise mean (should only be used if you know what you are doing).",
    )

    parser.add_argument(
        "--use-best-multisummit-IDR",
        default=False,
        action="store_true",
        help="Set the IDR value for a group of multi summit peaks (a group of peaks with the same chr/start/stop but different summits) to the best value across all of these peaks. This is a work around for peak callers that don't do a good job splitting scores across multi summit peaks (e.g. MACS). If set in conjunction with --plot two plots will be created - one with alternate summits and one without.  Use this option with care.",
    )

    parser.add_argument(
        "--allow-negative-scores",
        default=False,
        action="store_true",
        help="Allow negative values for scores. (should only be used if you know what you are doing)",
    )

    parser.add_argument(
        "--random-seed", type=int, default=0, help="The random seed value (sor braking ties). Default: 0"
    )
    parser.add_argument(
        "--max-iter",
        type=int,
        default=idr.MAX_ITER_DEFAULT,
        help="The maximum number of optimization iterations. Default: %i" % idr.MAX_ITER_DEFAULT,
    )
    parser.add_argument(
        "--convergence-eps",
        type=float,
        default=idr.CONVERGENCE_EPS_DEFAULT,
        help="The maximum change in parameter value changes "
        + "for convergence. Default: %.2e" % idr.CONVERGENCE_EPS_DEFAULT,
    )

    parser.add_argument("--only-merge-peaks", action="store_true", help="Only return the merged peak list.")

    parser.add_argument("--verbose", action="store_true", default=False, help="Print out additional debug information")
    parser.add_argument("--quiet", action="store_true", default=False, help="Don't print any status messages")

    parser.add_argument("--version", action="version", version="IDR %s" % idr.__version__)

    args = parser.parse_args()

    args.output_file = open(args.output_file, "w")
    idr.log_ofp = args.log_output_file

    if args.output_file_type is None:
        if args.input_file_type in ("narrowPeak", "broadPeak", "bed"):
            args.output_file_type = args.input_file_type
        else:
            args.output_file_type = "bed"

    if args.verbose:
        idr.VERBOSE = True

    global QUIET
    if args.quiet:
        idr.QUIET = True
        idr.VERBOSE = False

    if args.dont_filter_peaks_below_noise_mean is True:
        idr.FILTER_PEAKS_BELOW_NOISE_MEAN = False

    if args.allow_negative_scores is True:
        idr.ONLY_ALLOW_NON_NEGATIVE_VALUES = False

    assert idr.DEFAULT_IDR_THRESH == 1.0
    if args.idr_threshold == None and args.soft_idr_threshold == None:
        args.idr_threshold = idr.DEFAULT_IDR_THRESH
        args.soft_idr_threshold = idr.DEFAULT_SOFT_IDR_THRESH
    elif args.soft_idr_threshold == None:
        assert args.idr_threshold != None
        args.soft_idr_threshold = args.idr_threshold
    elif args.idr_threshold == None:
        assert args.soft_idr_threshold != None
        args.idr_threshold = idr.DEFAULT_IDR_THRESH

    numpy.random.seed(args.random_seed)

    if args.plot:
        try:
            import matplotlib
        except ImportError:
            idr.log(
                "WARNING: matplotlib does not appear to be installed and "
                + "is required for plotting - turning plotting off.",
                level="WARNING",
            )
            args.plot = False

    return args


def load_samples(args):
    # decide what aggregation function to use for peaks that need to be merged
    idr.log("Loading the peak files", "VERBOSE")
    if args.input_file_type in ["narrowPeak", "broadPeak"]:
        if args.rank == None:
            signal_type = "signal.value"
        else:
            signal_type = args.rank

        try:
            signal_index = {"score": 4, "signal.value": 6, "p.value": 7, "q.value": 8}[signal_type]
        except KeyError:
            raise ValueError("Unrecognized signal type for {} filetype: '{}'".format(args.input_file_type, signal_type))

        if args.peak_merge_method != None:
            peak_merge_fn = {"sum": sum, "avg": mean, "min": min, "max": max}[args.peak_merge_method]
        elif signal_index in (4, 6):
            peak_merge_fn = sum
        else:
            peak_merge_fn = min
        if args.input_file_type == "narrowPeak":
            summit_index = 9
        else:
            summit_index = None
        f1, f2 = [load_bed(fp, signal_index, summit_index) for fp in args.samples]
        oracle_pks = load_bed(args.peak_list, signal_index, summit_index) if args.peak_list != None else None
    elif args.input_file_type in [
        "bed",
    ]:
        # set the default
        if args.rank == None:
            signal_type = "score"

        if args.rank == "score":
            signal_type = "score"
            signal_index = 4
        else:
            try:
                signal_index = int(args.rank) - 1
                signal_type = "COL%i" % (signal_index + 1)
            except ValueError:
                raise ValueError(
                    "For bed files --signal-type must either "
                    + "be set to score or an index specifying "
                    + "the column to use."
                )

        if args.peak_merge_method == None:
            peak_merge_fn = sum
        else:
            peak_merge_fn = {"sum": sum, "avg": mean, "min": min, "max": max}[args.peak_merge_method]

        f1, f2 = [load_bed(fp, signal_index) for fp in args.samples]
        oracle_pks = load_bed(args.peak_list, signal_index) if args.peak_list != None else None
    elif args.input_file_type in [
        "gff",
    ]:
        # set the default
        if args.rank == None:
            signal_type = "score"
        else:
            assert args.rank == "score"

        if args.peak_merge_method == None:
            peak_merge_fn = sum
        else:
            peak_merge_fn = {"sum": sum, "avg": mean, "min": min, "max": max}[args.peak_merge_method]

        f1, f2 = [load_gff(fp) for fp in args.samples]
        oracle_pks = load_gff(args.peak_list) if args.peak_list != None else None
    else:
        raise ValueError("Unrecognized file type: '{}'".format(args.input_file_type))
    # build a unified peak set
    idr.log("Merging peaks", "VERBOSE")
    merged_peaks = merge_peaks([f1, f2], peak_merge_fn, oracle_pks, args.use_nonoverlapping_peaks)
    return merged_peaks, signal_type


def plot(args, scores, ranks, IDRs, ofprefix=None):
    assert len(args.samples) == 2
    import matplotlib

    matplotlib.use("Agg")
    import matplotlib.pyplot

    if ofprefix is None:
        ofprefix = args.output_file.name

    colors = numpy.zeros(len(ranks[0]), dtype=str)
    colors[:] = "k"
    colors[IDRs > args.soft_idr_threshold] = "r"

    # matplotlib.rc('font', family='normal', weight='bold', size=10)

    fig = matplotlib.pyplot.figure(num=None, figsize=(12, 12))

    matplotlib.pyplot.subplot(221)
    matplotlib.pyplot.axis([0, 1, 0, 1])
    matplotlib.pyplot.xlabel("Sample 1 Rank")
    matplotlib.pyplot.ylabel("Sample 2 Rank")
    matplotlib.pyplot.title("Ranks - (red >= %.2f IDR)" % args.soft_idr_threshold)

    try:
        matplotlib.pyplot.scatter(
            (ranks[0] + 1) / float(max(ranks[0]) + 1),
            (ranks[1] + 1) / float(max(ranks[1]) + 1),
            edgecolor=colors,
            c=colors,
            alpha=0.05,
        )
    except ValueError:  # empty ranks
        pass

    try:
        matplotlib.pyplot.subplot(222)
        matplotlib.pyplot.xlabel("Sample 1 log10 Score")
        matplotlib.pyplot.ylabel("Sample 2 log10 Score")
        matplotlib.pyplot.title("Log10 Scores - (red >= %.2f IDR)" % args.soft_idr_threshold)
        matplotlib.pyplot.scatter(
            numpy.log10(scores[0] + 1), numpy.log10(scores[1] + 1), edgecolor=colors, c=colors, alpha=0.05
        )
    except:
        pass

    def make_boxplots(sample_i):
        groups = defaultdict(list)
        norm_ranks = (ranks[sample_i] + 1) / float(max(ranks[sample_i]) + 1)
        for rank, idr_val in zip(norm_ranks, -numpy.log10(IDRs)):
            groups[int(20 * rank)].append(float(idr_val))
        group_labels = sorted((x + 2.5) / 20 for x in groups.keys())
        groups = [x[1] for x in sorted(groups.items())]

        matplotlib.pyplot.title("Sample %i Ranks vs IDR Values" % (sample_i + 1))
        matplotlib.pyplot.axis([0, 21, 0, 0.5 - math.log10(idr.CONVERGENCE_EPS_DEFAULT)])
        matplotlib.pyplot.xlabel("Sample %i Peak Rank" % (sample_i + 1))
        matplotlib.pyplot.ylabel("-log10 IDR")
        matplotlib.pyplot.xticks([], [])

        matplotlib.pyplot.boxplot(groups, sym="")

        matplotlib.pyplot.axis([0, 21, 0, 0.5 - math.log10(idr.CONVERGENCE_EPS_DEFAULT)])
        matplotlib.pyplot.scatter(20 * norm_ranks, -numpy.log10(IDRs), alpha=0.01, c="black")

    matplotlib.pyplot.subplot(223)
    try:
        make_boxplots(0)
    except ValueError:
        pass

    matplotlib.pyplot.subplot(224)
    try:
        make_boxplots(1)
    except ValueError:
        pass

    matplotlib.pyplot.savefig(ofprefix + ".png")
    return


def main():
    args = parse_args()

    # load and merge peaks
    merged_peaks, signal_type = load_samples(args)
    s1 = numpy.array([pk.signals[0] for pk in merged_peaks])
    s2 = numpy.array([pk.signals[1] for pk in merged_peaks])

    # build the ranks vector
    idr.log("Ranking peaks", "VERBOSE")
    r1, r2 = build_rank_vectors(merged_peaks)

    if args.only_merge_peaks:
        localIDRs, IDRs = None, None
    else:
        if len(merged_peaks) < -15:
            error_msg = "Peak files must contain at least 20 peaks post-merge"
            error_msg += "\nHint: Merged peaks were written to the output file"
            write_results_to_file(merged_peaks, args.output_file, args.output_file_type, signal_type)
            raise ValueError(error_msg)

        localIDRs = fit_model_and_calc_local_idr(
            r1,
            r2,
            starting_point=(args.initial_mu, args.initial_sigma, args.initial_rho, args.initial_mix_param),
            max_iter=args.max_iter,
            convergence_eps=args.convergence_eps,
            fix_mu=args.fix_mu,
            fix_sigma=args.fix_sigma,
        )

        # if the use chose to use the best multi summit IDR, then
        # make the correction and plot just the corrected peaks
        if args.use_best_multisummit_IDR:
            update_indices, localIDRs = correct_multi_summit_peak_IDR_values(localIDRs, merged_peaks)
            IDRs = calc_global_IDR(localIDRs)
            if args.plot:
                assert len(args.samples) == 2
                plot(
                    args,
                    [s1[update_indices], s2[update_indices]],
                    [r1[update_indices], r2[update_indices]],
                    IDRs[update_indices],
                    args.output_file.name + ".noalternatesummitpeaks",
                )
        # we wrap this in an else statement to avoid calculating the global IDRs
        # twice
        else:
            IDRs = calc_global_IDR(localIDRs)

        if args.plot:
            assert len(args.samples) == 2
            plot(args, [s1, s2], [r1, r2], IDRs)

    num_peaks_passing_thresh = write_results_to_file(
        merged_peaks,
        args.output_file,
        args.output_file_type,
        signal_type,
        localIDRs=localIDRs,
        IDRs=IDRs,
        max_allowed_idr=args.idr_threshold,
        soft_max_allowed_idr=args.soft_idr_threshold,
        useBackwardsCompatibleOutput=args.use_old_output_format,
    )

    args.output_file.close()


if __name__ == "__main__":
    try:
        main()
    finally:
        if idr.log_ofp != sys.stderr:
            idr.log_ofp.close()



#!/usr/bin/env python

""" MultiQC module to parse output from sequana"""
import os
import re
from collections import OrderedDict

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

# Initialise the logger
log = logging.getLogger("multiqc.sequana/bamtools_stats")


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/bamtools_stats",  # name that appears at the top
            anchor="sequana",  #
            target="sequana",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="pipelines multi Summary",
        )

        self.sequana_data = {}
        for myfile in self.find_log_files("sequana_bamtools_stats"):
            logging.info("Parsing {}".format(myfile))
            # print( myfile['f'] )       # File contents
            # print( myfile['s_name'] )  # Sample name (from cleaned filename)
            # print( myfile['fn'] )      # Filename
            # print( myfile['root'] )    # Directory file was in
            name = myfile["s_name"]
            if name.startswith("sequana_bamtools_stats_"):
                # name = "bamtools.stats." + name.replace("sequana_bamtools_stats_", "")
                name = name.replace("sequana_bamtools_stats_", "")
            self.sequana_data[name] = self.parse_logs(myfile["f"])

        """info = "<ul>"
        for this in sorted(self.sequana_data.keys()):
            info += '<li><a href="{}/summary.html">{}</a></li>'.format(this,this,this)
        info += "</ul>"
        href="http://sequana.readthedocs.io/en/main/"
        target = "Sequana"
        mname = '<a href="{}" target="_blank">{}</a> individual report pages:'.format(href, target)
        self.intro = '<p>{} {}</p>'.format( mname, info)
        """

        if len(self.sequana_data) == 0:
            log.debug("No samples found: sequana_bamtools_stats")
            raise UserWarning

        log.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_total_read_section()
        self.add_mapped_reads_section()
        self.add_duplicates_section()
        self.add_forward_reverse_section()
        self.add_failed_QC()
        self.add_mapped_vs_unmapped_chart()

    def parse_logs(self, log_dict):
        """Parse this kind of logs::

        **********************************************
        Stats for BAM file(s):
        **********************************************

        Total reads:       496
        Mapped reads:      491  (98.9919%)
        Forward strand:    235  (47.379%)
        Reverse strand:    261  (52.621%)
        Failed QC:         0    (0%)
        Duplicates:        0    (0%)
        Paired-end reads:  0    (0%)
        """
        # FIXME will fail if instead of 0.001% we have 1e-3%
        data = {}
        regexes = {
            "total_reads": r"Total reads:\s*(\d+)",
            "mapped_reads": r"Mapped reads:\s*(\d+)",
            "mapped_reads_pct": r"Mapped reads:\s*\d+\s+\((.+)%\)",
            "forward_strand": r"Forward strand:\s*(\d+)",
            "forward_strand_pct": r"Forward strand:\s*\d+\s+\((.+)%\)",
            "reverse_strand": r"Reverse strand:\s*(\d+)",
            "reverse_strand_pct": r"Reverse strand:\s*\d+\s+\((.+)%\)",
            "failed_qc": r"Failed QC:\s*(\d+)",
            "failed_qc_pct": r"Failed QC:\s*\d+\s+\((.+)%\)",
            "duplicates": r"Duplicates:\s*(\d+)",
            "duplicates_pct": r"Duplicates:\s*\d+\s+\((.+)%\)",
            "paired_end": r"Paired-end reads:\s*(\d+)",
            "paired_end_pct": r"Paired-end reads:\s*\d+\s+\((.+)%\)",
            "proper_pairs": r"'Proper-pairs'\s*(\d+)",
            "proper_pairs_pct": r"'Proper-pairs'\s*\d+\s+\((.+)%\)",
            "both_mapped": r"Both pairs mapped:\s*(\d+)",
            "both_mapped_pct": r"Both pairs mapped:\s*\d+\s+\((.+)%\)",
            "read_1": r"Read 1:\s*(\d+)",
            "read_2": r"Read 2:\s*(\d+)",
            "singletons": r"Singletons:\s*(\d+)",
            "singletons_pct": r"Singletons:\s*\d+\s+\((.+)%\)",
        }

        import re

        for k, v in regexes.items():
            res = re.findall(v, log_dict)
            if len(res) == 0:
                # logging.warning("Found no entries for {}".format(k))
                data[k] = None
            elif len(res) == 1:
                data[k] = float(res[0])
            else:
                logging.warning("Found several entries for {}".format(k))
                data[k] = None
        return data

    def add_failed_QC(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"failed_qc_pct": self.sequana_data[name]["failed_qc_pct"]}

        pconfig = {
            "title": "Failed QC (%)",
            "percentages": True,
            "min": 0,
            "max": 100,
            "logswitch": False,
        }

        total = sum([data[name]["failed_qc_pct"] for name in self.sequana_data.keys()])
        if total == 0:
            plot = None
            description = "Failed QC (none found)."
        else:
            plot = bargraph.plot(data, None, pconfig)
            description = "Failed QC (%)."

        self.add_section(name="Failed QC (%)", anchor="failed_qc", description="Failed QC", helptext="", plot=plot)

    def add_forward_reverse_section(self):
        data = {}
        for name in self.sequana_data.keys():
            forward = self.sequana_data[name]["forward_strand_pct"]
            reverse = self.sequana_data[name]["reverse_strand_pct"]
            if reverse is None:
                reverse = 0
            if forward is None:
                forward = 0
            data[name] = {"fwd": forward, "rev": reverse}

        pconfig = {
            "title": "Forward/reverse",
            "percentages": True,
            "min": 0,
            "max": 100,
            "logswitch": False,
        }

        keys = OrderedDict()
        keys["fwd"] = {"color": "#437bb1", "name": "Forward"}
        keys["rev"] = {"color": "#b1084c", "name": "Reverse"}

        self.add_section(
            name="Forward/Reverse",
            anchor="fwd_rev",
            description="Forward reverse",
            helptext="",
            plot=bargraph.plot(data, keys, pconfig),
        )

    def add_total_read_section(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"total_read": self.sequana_data[name]["total_reads"]}

        pconfig = {
            "title": "Total reads",
            "percentages": False,
            "min": 0,
            "logswitch": True,
        }

        self.add_section(
            name="Total reads",
            anchor="total_read",
            description="Total reads found in the BAM file.",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_mapped_reads_section(self):
        data = {}
        for name in self.sequana_data.keys():
            mapped_reads = self.sequana_data[name]["mapped_reads_pct"]
            data[name] = {"mapped_reads_pct": mapped_reads}
        pconfig = {
            "title": "Mapped reads (%)",
            "percentages": False,
            "min": 0,
            "max": 100,
            "logswitch": True,
        }

        self.add_section(
            name="Mapped reads (%)",
            anchor="mapped reads",
            description="Total mapped reads found in the BAM file.",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_duplicates_section(self):
        data = {}
        for name in self.sequana_data.keys():
            duplicates_pct = float(self.sequana_data[name]["duplicates_pct"])
            data[name] = {"duplicates_pct": duplicates_pct}

        pconfig = {
            "title": "% Duplicates (%)",
            "description": "% Duplicated Reads",
            "suffix": "%",
            "min": 0,
            "max": 100,
            "scale": "OrRd",
            "logswitch": False,
        }

        total = sum([data[name]["duplicates_pct"] for name in self.sequana_data.keys()])
        if total == 0:
            plot = None
            description = "Duplicated reads (none found)."
        else:
            plot = bargraph.plot(data, None, pconfig)
            description = "Duplicated reads."

        self.add_section(name="Duplicates (%)", anchor="duplicates", description=description, helptext="", plot=plot)

    def add_mapped_vs_unmapped_chart(self):
        data = {}
        for name in self.sequana_data.keys():
            total = float(self.sequana_data[name]["total_reads"])
            mapped = float(self.sequana_data[name]["mapped_reads"])
            unmapped = total - mapped
            data[name] = {"reads_mapped": 100 * mapped / total, "reads_unmapped": 100 * unmapped / total}

        pconfig = {
            "title": "Mapped vs unmapped reads (%)",
            "percentages": True,
            "min": 0,
            "max": 100,
            "logswitch": False,
        }

        keys = OrderedDict()
        keys["reads_mapped"] = {"color": "#437bb1", "name": "Mapped"}
        keys["reads_unmapped"] = {"color": "#b1084c", "name": "Unmapped"}

        self.add_section(
            name="Mapped vs unmapped reads (%)",
            anchor="bamtools_stats_alignment",
            description="Alignment metrics",
            plot=bargraph.plot(data, keys, pconfig),
        )

    def populate_columns(self):
        headers = {}

        if any(["total_reads" in self.sequana_data[s] for s in self.sequana_data]):
            headers["total_reads"] = {
                "title": "Read count",
                "description": "read count",
                "min": 0,
                "scale": "RdYlGn",
                "format": "{0:.0f}",
                "shared_key": "count",
            }

        if any(["total_reads" in self.sequana_data[s] for s in self.sequana_data]):
            headers["mapped_reads_pct"] = {
                "title": "Mapped reads (%)",
                "description": "mapped reads (%)",
                "min": 0,
                "max": 100,
                "scale": "RdYlGn",
                "format": "{0:.3f}",
                "shared_key": "count",
            }

        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)



#!/usr/bin/env python

""" MultiQC module to parse output from sequana"""
import os
import re

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

# Initialise the logger
import colorlog

log = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/isoseq_qc",  # name that appears at the top
            anchor="sequana_isoseq_qc",  # ??
            target="sequana_isoseq_qc",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="pipelines multi Summary",
        )

        self.sequana_data = {}
        for myfile in self.find_log_files("sequana_isoseq_qc"):
            name = myfile["s_name"]
            try:
                parsed_data = self.parse_logs(myfile["f"])
            except Exception as err:
                print("{} could not be parsed".format(name))
                print(err)
                continue
            name = parsed_data["s_name"]
            self.sequana_data[name] = parsed_data

        if len(self.sequana_data) == 0:
            log.debug("No samples found: sequana_isoseq_qc")
            raise UserWarning

        info = "<ul>"
        for this in sorted(self.sequana_data.keys()):
            info += '<li><a href="{}/summary.html">{}</a></li>'.format(this, this, this)
        info += "</ul>"
        href = "http://sequana.readthedocs.io/en/main/"
        target = "Sequana"
        mname = '<a href="{}" target="_blank">{}</a> individual report pages:'.format(href, target)
        self.intro = "<p>{} {}</p>".format(mname, info)

        log.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_productivity()
        self.add_total_bases()

    def add_total_bases(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {
                "total_bases_gb": self.sequana_data[name]["total_bases_gb"],
            }
        pconfig = {
            "title": "Total bases (Gb)",
            "percentages": False,
            "min": 100,
            "logswitch": True,
        }
        self.add_section(
            name="total bases (Gb)",
            anchor="total_bases_gb",
            description="",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_productivity(self):
        from collections import OrderedDict

        data = OrderedDict()
        for name in self.sequana_data.keys():
            data[name] = OrderedDict(
                {
                    "P0": self.sequana_data[name]["P0"],
                    "P1": self.sequana_data[name]["P1"],
                    "P2": self.sequana_data[name]["P2"],
                }
            )

        pconfig = {
            "title": "Productivity",
            "percentages": True,
            "min": 0,
            "max": 0,
            "logswitch": False,
        }
        self.add_section(
            name="Productivity",
            anchor="productivity",
            description="P0/P1/P2 productivities",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def parse_logs(self, log_dict):
        import json

        log_dict = json.loads(log_dict)
        data = {}
        data["s_name"] = log_dict["sample_name"]

        log_dict = log_dict["data"]["QC"]
        # data['count'] = log_dict['data']["count"]
        data["P0"] = log_dict["productivity"]["P0"]
        data["P1"] = log_dict["productivity"]["P1"]
        data["P2"] = log_dict["productivity"]["P2"]
        data["total_bases_gb"] = log_dict["total_bases_gb"]
        data["read_length_poly_AVG"] = log_dict["read_length"]["polymerase"]["AVG"]
        data["read_length_poly_N50"] = log_dict["read_length"]["polymerase"]["N50"]

        return data

    def populate_columns(self):
        from collections import OrderedDict

        headers = OrderedDict()

        if any(["total_bases_gb" in self.sequana_data[s] for s in self.sequana_data]):
            headers["total_bases_gb"] = {
                "title": "total bases (Gb)",
                "description": "Total bases (Gb)",
                "min": 0,
                "max": 100,
                "scale": "RdYlGn",
                "format": "{:,.0d}",
                "shared_key": "productivity",
            }

        for this in ["P0", "P1", "P2"]:
            if any([this in self.sequana_data[s] for s in self.sequana_data]):
                headers[this] = {
                    "title": this,
                    "description": this,
                    "min": 0,
                    "max": 100,
                    "scale": "RdYlGn",
                    "format": "{:,.0d}",
                    "shared_key": "productivity",
                }

        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)



#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Environments manager"""
import math
import os
import pathlib
import shutil
import sys
import tarfile

from tqdm import tqdm

from damona.common import BinaryReader, Damona

manager = Damona()


import colorlog

logger = colorlog.getLogger(__name__)


__all__ = ["Environ", "Environment"]


class Environment:
    """Class to handle a specific environment given its name

    ::

        from damona import Environment
        ee = Environment("test1")
        ee.create_bundle("test.tar")
        ee.create_yaml("test.yaml")

    """

    def __init__(self, name):
        """.. rubric:: **Constructor**

        :param str name: the name of the environment
        """
        self._init(name)

    def _init(self, name):
        self.name = name
        self.path = manager.damona_path / f"envs/{name}/"

        if self.path.exists() is False:
            logger.error(f"Environment {self.path} does not exists")
            sys.exit(1)

    def get_installed_binaries(self):
        """Return all binaries of the environment"""
        binaries = self.path.glob("bin/*")
        binaries = [x for x in binaries if x.is_dir() is False]

        return [x for x in binaries]

    def __contains__(self, name):
        binaries = [x.name for x in self.get_installed_binaries()]
        return name in binaries

    def get_disk_usage(self):
        """Return virtual size of the environment if we were to
        copy/export all images"""
        S = 0
        images = self.get_images()
        for image in images:
            try:
                if "DAMONA_PATH" not in os.environ:  # pragma: no cover
                    logger.error("You must define a DAMONA_PATH")
                    sys.exit(1)
                damona_path = os.environ["DAMONA_PATH"]
                image = image.replace("${DAMONA_PATH}", damona_path)
                S += os.path.getsize(image)
            except Exception as err:  # pragma: no cover
                print(err)
                logger.error(f"Could not check {image}")
        return S

    def get_images(self):
        """Return list of singularity images used by the environment"""
        binaries = self.get_installed_binaries()
        images = []
        for binary in binaries:
            br = BinaryReader(binary)
            if br.image not in images:
                images.append(br.image)
        return images

    def __repr__(self):
        N = len(self.get_installed_binaries())
        M = len(self.get_images())

        txt = f"Contains {N} binaries from {M} images."
        disk = self.get_disk_usage()
        import math

        env_size = math.ceil(disk / 1e6)

        txt += f" Disk usage is {env_size}Mb"
        return txt

    def rename(self, newname, force=False):
        """Rename an environment.

        Note that the *base* environment cannot be renamed
        """
        if self.name == "base":
            logger.error("You cannot rename the 'base' environment")
            sys.exit(1)

        if newname in Environ().environment_names:
            logger.error(f"{newname} exists already. Please choose another name")
            sys.exit(1)

        if not force:
            logger.warning(f"You are about to rename your current {self.name} environment into {newname}")
            input("Press enter to accept this change")

        ff = pathlib.Path(self.path)
        ff.rename(self.path.parent / newname)
        # reset the path
        self._init(newname)
        logger.warning("Please restart a new shell if this is an active environment.")

    def get_current_state(self):
        """Return dictionary with statistics about the environment

        It includes the number of binariesm images and name of the
        environment for now.
        """
        images = set()
        binaries = {}
        for filename in self.get_installed_binaries():
            br = BinaryReader(filename)
            image = br.get_image()
            binaries[filename.name] = image
            images.add(image)
        return {"images": images, "binaries": binaries, "name": self.name}

    def create_bundle(self, output_name=None):
        """Create a bundle with all images and binaries used by the environment.

        :param str output_name: if provided this will be the output filename
        :return: the name of the bundle. If output_name is None, set to damona_ENVNAME.tar
        """
        if output_name is None:
            output_name = f"damona_{self.name}.tar"

        # for later maybe
        exclude = []

        # all binaries
        binaries = self.get_installed_binaries()
        binaries = [x.absolute() for x in binaries if x not in exclude]
        binaries = sorted(binaries)

        # all containers
        images = [pathlib.Path(x) for x in self.get_images()]

        archive = tarfile.open(output_name, "w")
        for filename in binaries:
            logger.info(f"Adding {filename}")
            archive.add(filename, arcname=f"bin/{filename.name}")
        for filename in images:
            logger.info(f"Adding {filename}")
            archive.add(filename, arcname=f"images/{filename.name}")
        archive.close()

        logger.info(f"Saved environment {self.name} into {output_name}")
        return output_name

    def create_yaml(self, output_name=None):
        if output_name is None:
            output_name = f"damona_{self.name}.yaml"

        images = [pathlib.Path(x) for x in self.get_images()]

        with open(output_name, "w") as fout:
            fout.write(f"name: {self.name}\n")
            fout.write(f"\nimages:\n")

            for image in images:
                fout.write(f"- {image.name}\n")

            fout.write(f"\nbinaries:\n")

            binaries = self.get_installed_binaries()
            binaries = [x.absolute() for x in binaries]
            binaries = sorted(binaries)

            for binary in binaries:
                bininst = BinaryReader(binary)
                fout.write(f"- {binary.name} from {bininst.get_image()}\n")


class YamlEnv:
    def __init__(self, filename):
        self.name = None
        self.binaries = []
        self.images = []

        with open(filename, "r") as fin:
            for line in fin.readlines():
                if line.startswith("name:"):
                    self.name = line.split(":")[1].strip()

                elif line.startswith("-") and " from " in line:
                    line = line.replace("- ", "").strip()
                    self.binaries.append(line)
                elif line.startswith("-") and "from" not in line:
                    line = line.replace("- ", "").strip()
                    self.images.append(line)


class Images:
    def __init__(self):
        self.images_dir = manager.images_directory

    def __len__(self):
        return len(list(self.files))

    def _get_images(self):
        return self.images_dir.glob("*.img")

    files = property(_get_images)

    def get_disk_usage(self, frmt="Mb"):
        env_size = sum(os.path.getsize(f) for f in self.files if os.path.isfile(f))
        if frmt == "Mb":
            return math.ceil(env_size / 1e6)
        else:
            return env_size


class Environ:
    """Class to deal with the damona environments"""

    def __init__(self):
        self.images = Images()

    @staticmethod
    def get_current_env():
        if "DAMONA_ENV" not in os.environ:  # pragma: no cover
            logger.error(
                "You do not have any environment activated. Please use "
                "'damona activate ENVNAME' where ENVNAME is a valid environment"
            )
            sys.exit(1)
        else:
            return pathlib.Path(os.environ["DAMONA_ENV"])

    @staticmethod
    def get_current_env_name(warning=True):
        if "DAMONA_ENV" not in os.environ:
            if warning:
                logger.warning(
                    "You do not have any environment activated. Please use "
                    "'damona activate ENVNAME' where ENVNAME is a valid environment"
                )
            return None
        else:
            path = pathlib.Path(os.environ["DAMONA_ENV"])
            return path.name

    def _get_N(self):
        return len(self.environments)

    N = property(_get_N)

    def _get_envs(self):
        path_envs = manager.damona_path / "envs"
        envs = [Environment(x.name) for x in path_envs.iterdir()]
        return envs

    environments = property(_get_envs)

    def _get_env_names(self):
        envs = os.listdir(manager.environments_path)
        envs = sorted([Environment(x).name for x in envs])
        return envs

    environment_names = property(_get_env_names)

    def delete(self, env_name, force=False):
        if env_name == "base":
            logger.error("Environment 'base' is reserved and cannot not be created or deleted")
            sys.exit(1)

        env_path = manager.environments_path / env_name
        if os.path.exists(env_path) is False:
            logger.error("{} does not exists".format(env_path))
        else:
            try:
                os.rmdir(env_path)
            except OSError:
                if not force:
                    logger.warning(
                        f"Will delete all contents of {env_path}. Although this concerns only aliases you will lose your environement"
                    )
                else:
                    logger.warning(f"Deleting environment {env_path} since you used --force")

                if force:
                    shutil.rmtree(env_path)
                else:
                    ret = input("Are you sure you want to proceed ? (N/y)")
                    if ret == "y":
                        shutil.rmtree(env_path)

    def _env_in_path(self, env_name):
        PATH = os.environ["PATH"]
        paths = PATH.split(":")
        for x in paths:
            if str(manager.damona_path / "envs" / env_name / "bin") == x:
                return True
        return False

    def _is_fish_shell(self):
        if "DAMONA_SHELL_INFO" in os.environ:
            return os.environ["DAMONA_SHELL_INFO"] == "fish"
        return False

    def _is_bash_shell(self):
        if "DAMONA_SHELL_INFO" in os.environ:
            return os.environ["DAMONA_SHELL_INFO"] == "bash"
        return False

    def _is_zsh_shell(self):
        if "DAMONA_SHELL_INFO" in os.environ:
            return os.environ["DAMONA_SHELL_INFO"] == "zsh"
        return False

    def activate(self, env_name=None):
        # Do not change the print statement here below. They are used by
        # damona.sh
        if env_name not in self.environment_names:
            logger.error(f"invalid environment:  {env_name}. Please use 'damona env' to get the list")
            sys.exit(1)

        if self._env_in_path(env_name) is True:
            logger.warning(f"damona environment {env_name} is already in your PATH. nothing done")
            return

        env_path = manager.environments_path / env_name

        if self._is_fish_shell():
            print(f"    set -gx DAMONA_ENV {env_path};")
            print(f"    set -gx fish_user_paths {env_path}/bin $fish_user_paths")
        elif self._is_bash_shell():
            print(f"   export DAMONA_ENV={env_path};")
            print("    export PATH={}/bin:${{PATH}}".format(env_path))
        elif self._is_zsh_shell():
            print(f"   export DAMONA_ENV={env_path};")
            print("    export PATH={}/bin:${{PATH}}".format(env_path))
        else:
            _shell_info = os.environ["DAMONA_SHELL_INFO"]
            raise NotImplementedError(f"shell info found: {_shell_info}")
        logger.info(f"# Added damona path ({env_path}) in your PATH")

    def deactivate(self, env_name=None):
        # we deactivate the latest activated damona environment only.
        # can be called several times. If called too many times,
        # we set the main damona environment (base) as default
        PATH = os.environ["PATH"]
        paths = PATH.split(":")

        found = False  # this one is the one to deactivate (to ignore)
        newPATH = []
        for path in paths:
            # if deactivate without name, we remove the last one only
            if env_name and str(manager.damona_path / "envs" / env_name / "bin") == path:
                logger.info(f"# Found damona path ({path}), to be removed from your PATH")
                found = True
            elif not env_name and "/damona/envs/" in str(path) and not found:
                logger.info(f"# Found a damona path ({path}), to be removed from your PATH")
                found = True
            else:  # keep track of other paths.
                newPATH.append(path)

        if found is False:
            logger.info("# no more active damona environment in your path. Use 'damona activate ENVNAME'")

        damona_paths = [x for x in newPATH if "/damona/envs/" in x]
        if len(damona_paths):
            # in theory, there must be a /bin at the end of the path;
            # we should get rid of it
            first_damona_path = pathlib.Path(damona_paths[0])
            assert first_damona_path.name == "bin", "found a damona path with name different from 'bin'"
            first_damona_path = str(first_damona_path.parent)

            if self._is_fish_shell():
                print(f"set DAMONA_ENV {first_damona_path};")
            else:
                print(f"    export DAMONA_ENV={first_damona_path};")
        else:
            if self._is_fish_shell():
                print("set -e DAMONA_ENV;")
            else:
                print("    unset DAMONA_ENV")

        if self._is_fish_shell():
            print("set -e fish_user_paths;")
            for x in damona_paths:
                print(f"set -gx fish_user_paths {x} $fish_user_paths;")
        else:
            newPATH = ":".join(newPATH)
            print("export PATH={}".format(newPATH))

    def create(self, env_name, force=False):
        if env_name == "base":
            logger.critical("base is a reserved name for environment. Cannot be created")
            sys.exit(1)

        env_directory = manager.damona_path
        env_path = env_directory / "envs" / env_name
        if os.path.exists(env_path) and force is False:
            logger.critical("{} exists already".format(env_path))
            sys.exit(0)
        else:
            try:
                os.mkdir(env_path)
                os.mkdir(env_path / "bin")
                logger.info(f"Created {env_name} in {env_directory}")
                logger.info(f"Type 'damona activate {env_name}' to activate it")
            except:  # pragma: no cover
                pass  # if already created, error are caught here

    def create_from_yaml(self, env_name, yaml, force=False):
        if env_name in self.environment_names:
            logger.warning(f"{env_name} exists already.")
            if force is False:
                logger.critical(f"To recreate, you must delete it using 'damona delete {env_name}'")
                sys.exit(0)
            else:
                logger.warning("You used --force, so overwritting existing environment")

        env_directory = pathlib.Path(manager.damona_path / "envs" / env_name)

        from damona.environ import YamlEnv

        ye = YamlEnv(yaml)
        # ye.name is not used for now, supposibly, it already exists, so we require a user input

        self.create(env_name, force=force)

        for image in tqdm(ye.images):
            # local import to avoid circular import
            from damona.install import RemoteImageInstaller

            # replace all _ with :
            image = image.rsplit(".", 1)[0].replace("_", ":")

            # and replace : with _ except last one
            image = image.replace(":", "_", image.count(":") - 1)

            rii = RemoteImageInstaller(image)
            rii.pull_image(force=True)

        for binary in tqdm(ye.binaries):
            # local import to avoid circular import
            from damona.install import BinaryInstaller

            binary, image = binary.split(" from ")
            binary = binary.strip()
            image = image.strip()
            image += ".img"
            bi = BinaryInstaller([binary], image, env_name)
            bi.install_binaries()

    def create_from_bundle(self, env_name, bundle, force=False):
        if env_name in self.environment_names:
            logger.warning(f"{env_name} exists already.")
            if force is False:
                logger.critical(f"To recreate, you must delete it using 'damona delete {env_name}'")
                sys.exit(0)
            else:
                logger.warning("You used --force, so overwritting existing environment")

        # if it does not exsits or force is True
        self.create(env_name, force=force)

        env_directory = pathlib.Path(manager.damona_path / "envs" / env_name)

        archive = tarfile.open(bundle)
        for x in archive.getmembers():
            if x.name.startswith("bin"):  # this is a binary
                archive.extract(x, path=env_directory)
            elif x.name.startswith("images"):  # a contaier
                archive.extract(x, path=env_directory)
                from damona.common import ImageReader

                temp_image = ImageReader(env_directory / x.name)
                if temp_image.is_installed():
                    target_image = ImageReader(manager.damona_path / x.name)
                    if target_image.md5 == temp_image.md5:
                        logger.info(f"- Image {x.name} exists already with same md5sum. Skipped copy.")
                    else:
                        logger.warning(
                            f"- Image {x.name} exists already with diffrent "
                            " md5sum. This should not happen with damona images."
                            " You will need to copy manually if required"
                        )
                else:
                    logger.info(f"- Copying {x.name} into {manager.damona_path}/images")
                    temp_image.filename.rename(manager.damona_path / "images" / temp_image.filename.name)
            else:  # we should not enter here
                raise NotImplementedError

        # finally, we remove all images since they are redundant or have
        # been copied
        def rm_tree(pth):
            pth = pathlib.Path(pth)
            for child in pth.glob("*.img"):
                if child.is_file():
                    child.unlink()
                else:
                    rm_tree(child)
            pth.rmdir()

        to_delete = (env_directory / "images").absolute()
        logger.info(f"Removing temporary {to_delete} directory")
        rm_tree(env_directory / "images")

    def copy(self):
        """Copy an environment"""
        raise NotImplementedError



###########################################################################
# Damona is a project to manage reproducible containers                   #
#                                                                         #
# Authors: see CONTRIBUTORS.rst                                           #
# Copyright © 2020-2021  Institut Pasteur, Paris and CNRS.                #
# See the COPYRIGHT file for details                                      #
#                                                                         #
# Damona is free software: you can redistribute it and/or modify          #
# it under the terms of the GNU General Public License as published by    #
# the Free Software Foundation, either version 3 of the License, or       #
# (at your option) any later version.                                     #
#                                                                         #
# Damona  is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of          #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
# GNU General Public License for more details.                            #
#                                                                         #
# You should have received a copy of the GNU General Public License       #
# along with this program (COPYING file).                                 #
# If not, see <http://www.gnu.org/licenses/>.                             #
###########################################################################
""".. rubric:: Standalone application dedicated to Damona"""
import os
import pathlib
import sys
import time
import packaging
import subprocess

import click
import click_completion
import rich_click as click

click_completion.init()


from damona import Damona, Environ, Environment, version
from damona.common import BinaryReader, ImageReader, get_damona_path
from damona.install import (
    BiocontainersInstaller,
    LocalImageInstaller,
    RemoteImageInstaller,
)
from damona.registry import BiocontainersRegistry, Registry

click.rich_click.USE_MARKDOWN = True
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.STYLE_ERRORS_SUGGESTION = "magenta italic"
click.rich_click.SHOW_ARGUMENTS = True

manager = Damona()


__all__ = ["main", "build"]

from damona import logger

logger.level = 10


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--level",
    default="INFO",
    help="""Set level information to DEBUG, INFO, WARNING, CRITICAL, ERROR. Use e.g., "damona --level INFO command" """,
)
@click.version_option(version=version)
def main(level):
    """Damona is an environment manager for singularity containers.

    It is to singularity container what conda is to packaging.

    The default environment is called 'base'. You can create and activate
    a new environment as follows:

        damona create --name TEST
        damona activate TEST

    Once an environment is activated, you can install a Damona-registered image
    (and its registered binaries):

        damona install fastqc:0.11.9

    More information on https://damona.readthedocs.io.
    Please report issues on https://github.com/cokelaer/damona
    Contact: Thomas Cokelaer at pasteur dot fr

    "Make everything as simple as possible, but not simpler." -- Albert Einstein
    """
    ######################## !!!!!!!!!!!! ####################
    # this function cannot print anything because the damona
    # activate command prints bash commands read by damona.sh
    ######################## !!!!!!!!!!!! ####################
    logger.setLevel(level)


@main.command()
@click.argument("environment", required=True, type=click.STRING)
@click.option("--from-bundle", type=click.STRING, help="a bundle file create with 'damona export --bundle' command")
@click.option("--from-yaml", type=click.STRING, help="a yaml file create with 'damona export --yaml' command")
@click.option(
    "--force",
    is_flag=True,
    help="""When creating an environment with --from-bundle, rewrite binaries and
images even though the environment exists.""",
)
def create(**kwargs):
    """Create a new environment

    Here we create an environment called TEST:

        damona create TEST

    You can then activate it:

        damona activate TEST

    You can create an environment from a environment.yaml file that was created with the 'export --yaml' command
    or manually built using the following syntax:


        name: sequana_rnaseq

        images:
            - sequana_tools_0.14.5.img

        binaries:
            - bwa
            - samtools
            - bamtools

    """
    envs = Environ()
    if kwargs["from_bundle"]:
        envs.create_from_bundle(kwargs["environment"], bundle=kwargs["from_bundle"], force=kwargs["force"])
    elif kwargs["from_yaml"]:
        envs.create_from_yaml(kwargs["environment"], yaml=kwargs["from_yaml"], force=kwargs["force"])
    else:
        envs.create(kwargs["environment"])


@main.command()
@click.argument("environment", required=True, type=click.STRING)
@click.option("--force", is_flag=True, help="""When creating an environment with --from-bundle, rewrite binaries and""")
def delete(**kwargs):
    """Remove an environment"""
    env = Environ()
    env.delete(kwargs["environment"], force=kwargs["force"])


@main.command()
@click.argument("environment", required=True, type=click.STRING)
@click.option("--new-name", required=True, type=click.STRING, help="""new name of the environments""")
def rename(**kwargs):
    """Rename an existing environment"""
    env = Environment(kwargs["environment"])
    env.rename(kwargs["new_name"])


# =================================================================== env
@main.command()
def env(**kwargs):
    """List all environemnts with some stats.

    Print information about current environments:

        damona env


    """
    envs = Environ()

    click.secho(f"There are currently {envs.N} Damona environments:\n", bold=True)
    if envs.N != 0:
        for this in envs.environments:
            name = this.name
            click.echo(click.style(f"{name}", bold=True) + click.style(f" -  {this}"))
    current_env = envs.get_current_env_name()
    click.secho(f"\nYour current env is '{current_env}'.", bold=True)


# =================================================================== activate
@main.command()
@click.argument("name", required=True, type=click.STRING)
def activate(**kwargs):
    """Activate a damona environment.

    The main Damona environment can be activated using:

        damona activate base

    Then, activation of a specific environment is done as:

        damona activate my_favorite_env

    """
    # DO NOT PRINT ANYTHING HERE OTHERWISE YOU'LL BREAK
    # DAMONA BASH EXPORT.If yo do, use # as commented text
    envs = Environ()
    envs.activate(kwargs["name"])


# =================================================================== deactivate
@main.command()
@click.argument("name", required=False, type=click.STRING, default=None)
def deactivate(**kwargs):
    """Deactivate the current Damona environment.

    deactivate the current environment:

        damona deactivate

    """
    # DO NOT PRINT ANYTHING HERE OTHERWISE YOU'LL BREAK
    # DAMONA BASH EXPORT.If yo do, use # as commented text
    env = Environ()
    env.deactivate(kwargs["name"])


# =================================================================== install
@main.command()
@click.argument("image", required=True, type=click.STRING)
@click.option("--force-image", is_flag=True, help="Replaces existing image.")
@click.option("--force", is_flag=True, help="Replaces images and binaries.")
@click.option("--force-binaries", is_flag=True, help="Replace binaries.")
@click.option(
    "--url",
    help="""download image from a remote URL. The URL must
  contain a registry.txt as explained on https://damona.readthedocs.io""",


)
@click.option(
    "--binaries",
    default=None,
    help="""If not provided, we assume this is an executable singulatrity and its name is the binary name
    """,
)
def install(**kwargs):
    """Download and install an image and its binaries.

    An image has a name and a version and can be installed as::

        damona install NAME:version

    If the version is omitted, the latest version is installed. Therefore, the
    following two commands are equivalent if 0.11.9 is the latest version available::


        damona install fastqc
        damona install fastqc:0.11.9

    You may also install a local image. In such case, you must name your
    image with a version (e.g. fastqc_0.11.9.img). We assume that name of
    the image is the binary name, however, binaries can be set manually::


        damona install fastqc_0.4.2.img
        damona install test_0.4.2.img --binary fastqc

    By convention, images must follow the pattern NAME_[v]x.y.z[_info].img
    where the [v] character is optional and the [_info] is an optional _
    character followed by any text. Finally, the extension can be .img or .sif

    Images are installed in the ~/.config/damona/images directory. It is the
    DAMONA_PATH environment variable. Therefore, you can redefine this variable
    to install images elsewhere.

    You may have images online on a website. To install such images, use
    the --from-url (see developer guide for details). The binary will be named
    after the name provided. For instance, this command download the image
    and creates a binary called 'fastqc'. ::

        damona install fastqc --from-url https://biomics.pasteur.fr/salsa/damona/fastqc_0.11.8.img


    Or wish to use an existing docker file::

        damona install biocontainers/hisat2:v2.0.5-1-deb

    Note (June 2023) biocontainers do not work anymore...

    """
    logger.debug(kwargs)

    env = Environ()
    cenv = env.get_current_env()

    image_path = pathlib.Path(kwargs["image"]).absolute()

    force_image = kwargs["force_image"]
    force_binaries = kwargs["force_binaries"]

    if kwargs["force"]:
        force_image, force_binaries = True, True

    if kwargs["binaries"]:
        if "," in kwargs["binaries"]:
            binaries = kwargs["binaries"].split(",")
        else:
            binaries = [kwargs["binaries"]]
    else:
        binaries = None

    if kwargs["image"].startswith("biocontainers/"):
        p = BiocontainersInstaller(kwargs["image"], binaries=binaries)
        p.pull_image(force=force_image)
        p.install_binaries(force=force_binaries)

    elif os.path.exists(image_path) is False or kwargs['url']:
        url = kwargs["url"]
        logger.info(f"Installing from given URL")

        p = RemoteImageInstaller(kwargs["image"], from_url=kwargs["url"], cmd=sys.argv, binaries=binaries)

        if p.is_valid():
            p.pull_image(force=force_image)
            p.install_binaries(force=force_binaries)
            with open(cenv / "history.log", "a+") as fout:
                cmd = " ".join(["damona"] + sys.argv[1:])
                fout.write(f"\n{time.asctime()}: {cmd}")
        else:
            logger.critical("Something wrong with your image/binaries. See message above")
            sys.exit(1)
    else:
        # This install the image and associated binary/binaries
        logger.info(f"Installing local container in {cenv}")

        lii = LocalImageInstaller(image_path, cmd=sys.argv, binaries=binaries)
        if lii.is_valid():
            lii.install_image(force=force_image)
            lii.install_binaries(force=force_binaries)
            with open(cenv / "history.log", "a+") as fout:
                cmd = " ".join(["damona"] + sys.argv[1:])
                fout.write(f"\n{time.asctime()}: {cmd}")
        else:
            logger.critical("Something wrong with your image/binaries. See message above")
            sys.exit(1)


# =================================================================== remove
@main.command()
@click.argument("name", required=True, type=click.STRING)
@click.option("--environment", type=click.STRING, default=None)
# @click.option("--force", is_flag=True, help="force the removal of binaries or images")
def remove(**kwargs):
    """Remove binaries or image from an environment.

    You can remove a binary from an environment given its path. It will not be removed
    if used by an executable in an environment.::

        damona remove /home/cokelaer/.config/damona/images/fastqc_0.11.8.img

    if you have the name of the image, it works as well::

        damona remove fastqc_0.11.8.img

    You must give the .img extension in both cases otherwise it is considered
    to be a binary that you want to remove.

    If you suppress a binary like here::

        damona remove fastqc

    it removes the binary from the activate environment only. Then, if the image is now an
    orphan, it is also removed.


    You can also remove an image (and its binaries) from an environment. Note, however,
    that the image is not deleted if used in other environments.
    """
    # First, let us figure out the current or user-defined environment
    envs = Environ()
    env_name = kwargs["environment"]
    if not env_name:
        env_name = envs.get_current_env_name(warning=False)
        if env_name is None:
            logger.error(
                "You must activate a damina environment or use the --environment to define one where binary:image will be removed."
            )
            sys.exit(1)

    env = Environment(env_name)
    dam = Damona()

    # then, let us figure out what the user wants (remove a binary or image ?) and do it
    name = kwargs["name"]
    if kwargs["name"].endswith(".img"):
        # we delete the image if it is an orphan
        p = get_damona_path() / "images" / kwargs["name"]
        if p.exists():
            ir = ImageReader(p)
            ir.delete()
        else:
            logger.warning(f"input file {p} does not exists")

    else:
        # Search for the name in the installed binaries
        if name in env:
            logger.info(f"Removing binary {name}")
            binary = [x for x in env.get_installed_binaries() if x.name == name]
            binary = binary[0]
            br = BinaryReader(binary)
            br.image

            # keep this info before deleting the file
            image_name = br.get_image()

            # we now delete the executable
            binary.unlink()

            # and the image if required
            p = get_damona_path() / "images" / (image_name.replace(":", "_") + ".img")
            ir = ImageReader(p)
            ir.delete()
        else:
            logger.warning(f"{name} was not found in the environment {env_name}. Not removed")

    with open(env.path / "history.log", "a+") as fout:
        cmd = " ".join(["damona"] + sys.argv[1:])
        fout.write(f"\n{time.asctime()}: {cmd}")


# =================================================================== clean
@main.command()
@click.option("--remove", is_flag=True, help="remove the binary and image orphans in all environments ")
def clean(**kwargs):
    """Remove orphan images and binaries from all environments.

    This command finds images that have no associated binaries in any environment.
    This may happen with prior version of a given binary.

    Note that if you want to remove an environmnt, just delete it or use::

        conda env --delete NAME

    """
    logger.debug(kwargs)

    dmn = Damona()

    # First we deal with orphans from the binaries directories
    orphans = dmn.find_orphan_binaries()
    if len(orphans) == 0:
        # nothing to do
        logger.info("No binary orphan found")
    else:
        logger.info(f"Found {len(orphans)} binary orphans.")
        if kwargs["remove"]:
            for x in orphans:
                os.remove(os.path.expanduser(x))
                logger.info(f"Removed {x}")
        else:
            logger.warning("Please use --remove to confirm that you want to remove the orphans")

    # Second, we find images that have no more binaries
    orphans = dmn.find_orphan_images()
    if len(orphans) == 0:
        logger.info("No orphan images found")
    else:
        logger.info(f"Found {len(orphans)} image orphans.")

        if kwargs["remove"]:  # pragma: no cover
            for x in orphans:
                os.remove(os.path.expanduser(x))
                logger.info(f"Removed {x}")
        else:
            logger.warning("Please use --remove to confirm that you want to remove the orphans")


# =================================================================== search
@main.command()
@click.argument("pattern", required=True, type=click.STRING)
@click.option("--images-only", is_flag=True, default=False, help="Show images only")
@click.option("--include-biocontainers", is_flag=True, default=False, help="include also biocontainers hits")
@click.option("--binaries-only", is_flag=True, default=False, help="Show binaries only")
@click.option(
    "--url",
    help="""Set the online registry file to search for a
given container. See damona.readthedocs.io for in formation on how to write this
file . Example is available on https://biomics.pasteur.fr/salsa/damona/registry.txt""",
)
def search(**kwargs):
    """Search for a container or binary.

    By default, this command introspect the official Damona registry (based on registry files)::

        damona search fastqc

    If you want to list all software and their versions, just type::

        damona search "*"

    With fish shells, use::

        damona search '"*"'

    One can also search in an online registry::

        damona search fastqc --url https://biomics.pasteur.fr/salsa/damona/registry.txt

    You may define aliases to URLs in your ~/.config/damona/damona.cfg file to
    make it easier::

        damona search fastqc --url damona

    Although not recommended (not curated by Damona), you can also
    install any container available on biocontainer. To do som you first
    need to know the name and version:

        damona search fastqc --include-biocontainers


    """

    url = kwargs.get("url", None)

    if kwargs["pattern"] == "*":
        pattern = None
    else:
        pattern = kwargs["pattern"]

    registry = Registry(from_url=url)

    click.echo()

    recommended = None

    if not kwargs["binaries_only"]:
        click.echo(f"Pattern '{pattern}' found in these releases:")
        modules = registry.get_list(pattern=pattern)
        for mod in modules:
            name, version = mod.split(":")
            url = registry.registry[mod]._data[name]['releases'][version]['download']
            try:
                size = registry.registry[mod]._data[name]['releases'][version]['filesize']
                if size > 1e9:
                    size =  round(size/1e9,2)
                    size = f"{size}G"
                else:
                    size =  round(size/1e6,2)
                    size = f"{size}M"
            except Exception:
                logger.warning(f"{mod}. could not extract filesize")
                size = "-1"

            click.echo(f" - {mod} -- {url} -- {size}")
            if not recommended:
                recommended = mod
            else:
                recommended_version = recommended.split(":")[1]
                try:
                    if packaging.version.parse(version) > packaging.version.parse(recommended_version):
                        recommended = mod
                except packaging.version.InvalidVersion:
                    pass
    click.echo()
    if not kwargs["images_only"]:
        click.echo(f"Pattern '{pattern}' found as binaries:")
        modules = registry.get_binaries(pattern=pattern)
        for mod in sorted(modules.keys()):
            v = modules[mod]
            name, version = mod.split(":")
            url = registry.registry[mod]._data[name]['releases'][version]['download']
            try:
                size = registry.registry[mod]._data[name]['releases'][version]['filesize']
                if size > 1e9:
                    size =  round(size/1e9,2)
                    size = f"{size}G"
                else:
                    size =  round(size/1e6,2)
                    size = f"{size}M"
            except Exception:
                logger.warning(f"{mod}. could not extract filesize")
                size = "-1"

            click.echo(f" - {mod}: -- {url} -- {size}")

    if kwargs["include_biocontainers"]:
        click.echo("Searching biocontainers:")
        br = BiocontainersRegistry()
        for k, data in br.data.items():
            if pattern in k:
                click.echo(f" - {k}: ")
                for version, location in data["releases"].items():
                    install = f"(damona install biocontainers/{k}:{version})"
                    click.echo(f" -     {version}: {install} ")
            elif pattern == "*":
                click.echo(f" - {k}: {v}")

    if recommended:
        click.echo(f"\n\n \U00002139\U0000FE0F -- Recommended installation (latest version and dedicated container) -- \U00002139\U0000FE0F \n\n    damona install {recommended}\n")




# ============================================================  export
@main.command()
@click.argument("environment", required=True, type=click.STRING)
def info(**kwargs):
    """Print information about a given environment.

    The default environment is called 'base'.::

        damona info base
        damona info test1

    Images abd binaries available are shown
    """
    logger.debug(kwargs)
    envname = kwargs["environment"]

    manager = Environ()

    x = [ee for ee in manager.environments if ee.name == envname]
    if len(x) == 0:
        logger.error(f"Environment '{envname}' does not exist. Use 'damona env' to get the list")
        sys.exit(1)
    else:
        environ = x[0]
        click.echo(f"name: {envname}")
        click.echo("\nimages:")
        for item in sorted(environ.get_images()):
            click.echo(" - " + pathlib.Path(item).name)
        click.echo("\nbinaries:")
        for item in sorted(environ.get_installed_binaries()):
            click.echo(" - " + pathlib.Path(item).name)


# ============================================================  export
@main.command()
@click.argument("environment", required=True, type=click.STRING)
@click.option("--yaml", help="name of output file")
@click.option("--bundle", default=None, help="name of output file")
def export(**kwargs):
    """Create a bundle of a given environment.

    the following command copies all binaries from an environment and their
    associated images into a tar ball file named after the
    environment.


        damona export TEST --yaml   test_env.yaml
        damona export TEST --bundle test_bundle.tar

    We do not compress the tar file. The images are already compressed.

    This create a bundle named damona_test1.tar. You can then create a new
    environment starting from this bundle:


        damona env --create TEST1 --from-bundle test_bundle.tar
        damona env --create TEST1 --from-yaml   test_env.yaml


    """
    from damona import Environment

    logger.debug(kwargs)

    environment = kwargs["environment"]
    envname = kwargs["environment"]

    # TODO This should be based on the binaries of the environment, not the images
    # to do so, we'll need an installed.txt file

    env = Environment(envname)
    if kwargs["bundle"]:
        bundle_file = kwargs["bundle"]
        output = env.create_bundle(output_name=kwargs["bundle"])
        logger.info(
            f"Use this command to create a new environment: \n\n\tdamona create test1 --from-bundle {bundle_file}"
        )
    elif kwargs["yaml"]:
        yaml_file = kwargs["yaml"]
        env.create_yaml(output_name=yaml_file)
        logger.info(f"Use this command to create a new environment: \n\n\tdamona create test1 --from-yaml {yaml_file}")


# ============================================================  stats


@main.command()
@click.option("--include-biocontainers", is_flag=True, help="include also biocontainers (experimental)")
@click.option("--include-downloads", is_flag=True, help="include downloads")
def stats(**kwargs):
    """Get information about Damona images and binaries

    Just type::

        damona stats

    This will print the actual status of Damona with number of binaries/images.
    This will also give local information

    """
    from damona import admin

    admin.stats()
    if kwargs["include_biocontainers"]:
        admin.stats(True)

    if kwargs["include_downloads"]:
        click.echo("Detailled summary of downloads for each container:")
        from damona import admin, zenodo

        all_software = admin.get_software_names()
        N = 0
        for software in sorted(all_software):
            downloads = zenodo.get_stats_software(software)
            click.echo(f"{software}: {downloads}")
            try:
                N += downloads.replace(",", "")
            except AttributeError:
                N += downloads

        click.echo(f"Total: {N}")

    envs = Environ()
    N = len(envs.images)
    usage = envs.images.get_disk_usage()
    click.echo(
        f"\n--\nLocal installation. In your local environment, we found {N} images. This account for a total of: {usage}.Mb"
    )


# ===================================================================  list


@main.command()
def list(**kwargs):
    """List all packages that can be installed"""
    r = Registry()
    names = "\n".join(sorted([x for x in r.get_list()]))
    click.echo(names)


# ============================================================  HIDDEN commands
# ============================================================  zenodo-upload


@main.command(hidden=True)
@click.argument("filename", required=True)
@click.option(
    "--token",
    default=None,
    help="""A valid zenodo (or sandbox zenodo) token (see damona zenodo --help for details).""",
)
@click.option("--mode", default="sandbox.zenodo", help="mode can be either 'zenodo' or 'sandbox.zenodo'")
@click.option("--no-check", default=False, help="Damona is driven by the Sequana project. Presence of bash and python are usually required by Sequana pipelines (Snakemake) so we make them compulsary. If developers do not need them, the --no-check option may be used. "
)
def upload(**kwargs):  # pragma: no cover
    """Upload a singularity file to Zenodo. FOR DEVELOPERS ONLY

    This command is for developers of the DAMONA project only.

    The sandbox.zenodo is a sandbox where you can try to upload a new singularity file:

        damona upload file_1.0.0.img --mode sandbox.zenodo

    Once done and happy with the results, you can upload to Zenodo itself once and for all:

        damona upload file_2.0.0.img --mode sandbox.zenodo

    If no registry.yaml is found in the local directory, it is created.
    Otherwise, it is updated. The changes are also printed on the stdout.

    You can set the token in your home/.config/damona/damona.cfg that looks like

        [general]
        quiet=False

        [urls]
        damona=https://biomics.pasteur.fr/salsa/damona/registry.txt

        [zenodo]
        token=APmm6p...
        orcid=0000-0001-...
        affiliation=Your Institute
        name=Surname, firstname

        [sandbox.zenodo]
        token=FFmbAE...
        orcid=0000-0001-...
        affiliation=Your Institute
        name=Surname, firstname

    """
    from damona.zenodo import Zenodo

    # some aliases
    token = kwargs["token"]
    mode = kwargs["mode"]
    filename = kwargs["filename"]

    # check that python and bash are available in the container.
    status = subprocess.run(f"singularity exec {filename} python --version".split(),
        stdout=subprocess.PIPE)
    if status.returncode:
        click.echo("Damona Warning: could not find **python** command in the container")
        proceed = click.prompt("Do you want to proceed ?")
        if proceed:
            click.echo("Uploading without Python found in the container.")
        else:
            click.echo("Exiting...")
            sys.exit(1)

    status = subprocess.run(f"singularity exec {filename} bash --version".split(),
        stdout=subprocess.PIPE)
    if status.returncode:
        click.echo("Damona ERROR: could not find **bash** command in the container", 
            err=True)
        sys.exit(1)

    #
    z = Zenodo(mode, token)
    logger.info(f"Uploading to {mode}")
    z._upload(filename)


# =================================================================== build
@main.command(hidden=True)
@click.argument("filename", required=True, type=click.STRING)
@click.option("--destination", default=None, help="Not implemented yet")
@click.option("--force", is_flag=True, help="add --force option")
def build(**kwargs):  # pragma: no cover
    """Build a container from dockerhub, singularity file or damona recipes.

    Note that to use this command, you must have sudo permissions.
    If not, you will need to download existing images. See the "damona install
    command".

    You can build a singularity image from a local singularity file. Note that
    your Singulary recipes must follow the naming convention
    Singularity.NAME_x,y,z

    \b
        # a local recipes (recipes must have a version)
        build Singularity.salmon_1.3.0

    You may build an image from a singularity recipes to be found in Damona
    itself. In such case, the name and version are enough. Siuch recipes can be
    listed using "damona list"

        damona build salmon:1.3.0

    You may also build image from a docker image to be found on docker hub:

        damona build docker://biocontainers/bowtie2:v2.4.1_cv1

    If no version is added, you need to add one using --destination. This will fail:

        damona build docker://kapeel/hisat2

    so you need to informa damona about the version:

        damona build docker://kapeel/hisat2 --destination hisat2_v2.0.0.img


    """
    logger.debug(kwargs)
    filename = kwargs["filename"]
    force = kwargs["force"]
    destination = kwargs["destination"]

    if destination:
        if destination.endswith(".img") is False:
            logger.warning("You should end your image with the .img extension")
        if "_" not in destination:
            logger.warning("You should name your image as NAME_x.y.z")

    if os.path.exists(filename) and os.path.isdir(filename) is False:
        # TODO check that it starts with Singularity
        # local recipes ?
        from damona.builders import BuilderFromSingularityRecipe

        builder = BuilderFromSingularityRecipe()
        builder.build(filename, destination=destination, force=force)
    elif kwargs["filename"].startswith("docker://"):
        from damona.builders import BuilderFromDocker

        builder = BuilderFromDocker()
        filename = filename.replace("docker://", "")
        builder.build(filename, destination=destination, force=force)
    else:  # could be a damona recipes
        logger.info("Not a docker URL, nor a local file.")


if __name__ == "__main__":  # pragma: no cover
    main()



import os
import sys

from setuptools import find_packages, setup

_MAJOR = 0
_MINOR = 16
_MICRO = 0
version = "%d.%d.%d" % (_MAJOR, _MINOR, _MICRO)
release = "%d.%d" % (_MAJOR, _MINOR)


metainfo = {
    "authors": {"main": ("Thomas Cokelaer", "thomas.cokelaer@pasteur.fr")},
    "maintainer": {"main": ("Thomas Cokelaer", "thomas.cokelaer@pasteur.fr")},
    "version": version,
    "license": "new BSD",
    "download_url": "https://github.com/sequana/sequana/archive/{0}.tar.gz".format(version),
    "url": "http://github.com/sequana/sequana",
    "description": "A set of standalone application and pipelines dedicated to NGS (new generation sequencing) analysis",
    "platforms": ["Linux", "Unix", "MacOsX", "Windows"],
    "keywords": ["NGS", "snakemake"],
    "classifiers": [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
    ],
}


packages = find_packages()
packages = [this for this in packages if this.startswith("test.") is False]
packages = [this for this in packages if this not in ["test"]]

# load a common list of requirements
requirements = open("requirements.txt").read().split()


setup(
    name="sequana",
    version=version,
    maintainer=metainfo["authors"]["main"][0],
    maintainer_email=metainfo["authors"]["main"][1],
    author=metainfo["authors"]["main"][0],
    author_email=metainfo["authors"]["main"][1],
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    keywords=metainfo["keywords"],
    description=metainfo["description"],
    license=metainfo["license"],
    platforms=metainfo["platforms"],
    url=metainfo["url"],
    download_url=metainfo["download_url"],
    classifiers=metainfo["classifiers"],
    # package installation
    packages=packages,
    # pillow, sphinx-gallery and numpydoc are  for the doc only
    # mock is for the test only qtconsole is required by Sequanix
    install_requires=requirements,
    # specific packages for testing
    # This is recursive include of data files
    exclude_package_data={"": ["__pycache__"]},
    package_data={
        "": [
            "README.rst",
            "requirements*txt",
        ],
        "sequana.multiqc": ["*yaml"],
    },
    # these files do not need to be added in MANIFEST.in since there are python
    # packages that will be copied from sequana/ into sequana/
    # Note, however, that e.g. ./pipelines must be added
    zip_safe=False,
    # ext_modules = ext_modules,
    entry_points={
        "multiqc.modules.v1": [
            "sequana_pacbio_qc=sequana.multiqc.pacbio_qc:MultiqcModule",
            "sequana_quality_control=sequana.multiqc.quality_control:MultiqcModule",
            "sequana_coverage=sequana.multiqc.coverage:MultiqcModule",
            "sequana_isoseq=sequana.multiqc.isoseq:MultiqcModule",
            "sequana_isoseq_qc=sequana.multiqc.isoseq_qc:MultiqcModule",
            "sequana_bamtools_stats=sequana.multiqc.bamtools_stats:MultiqcModule",
            "sequana_laa=sequana.multiqc.laa:MultiqcModule",
            "sequana_kraken=sequana.multiqc.kraken:MultiqcModule",
            "sequana_pacbio_amplicon=sequana.multiqc.pacbio_amplicon:MultiqcModule",
        ],
        "multiqc.hooks.v1": [
            "before_config = sequana.multiqc.config:load_config",
        ],
    },
)
